"""
 These objects correspond to the structure of the configurations that will form nodes from dataset and collection
 representations.

  The language here is intended to be more general than SQL-specific language, since the intention here is to be able to
  reference many different kinds of data resources. If we think of a traditional SQL database as composed of
 Schemas, Tables, and Fields, in this nomenclature
 Schema == Dataset: any abstract data source that may contain multiple data sets
 Table == Collection:  a set of identifiable fields (e.g. rows of data)
 Field == Field : a single field in a collection, addressable by name

Some examples:
 For a mongo db, Dataset/Collection/Field == Database/collection/field.
 For an API dataset,  Dataset/Collection/Field == API/endpoint/json response member

 In any case we will form a graph of nodes, where each traversal_node corresponds to a single collection in a single dataset, and
 connections can be made between any field in any one dataset to any other (with the exception of self-references, which
 are currently disallowed in this implementation).

 Each of these objects is identified by a unique string key/name ,and so a collection, belonging to a Dataset,
 can be identified by the unique tuple

 (data dataset name, data set name),

 referred to as a Collection Address. Similarly each field has a field address of the form
 (data dataset name, data set name, field name)

  uniquely identifying it in the graph. Since connections are field address <---> field address, we store a set of references on each field of their destinations.


After specifications:
 Both collections and Datasets support the specification of "after" addresses, to support manually tweaked iteration order.
 A Dataset "after" points to another data dataset. It means that any collection in this traversal_node will not be traversed until every traversal_node
 in the "after traversal_node is traversed.

 A collection "after" points to another collection. It means that this collection will not be traversed until after the traversal_node specified
 in the after Collection Address list is traversed.

 If this reordering cannot be satisfied it is an error condition (unreachable traversal_node).


Field identities:
 Field identities are pointers to seed values. They indicate that this field can be populated with seed values, and so are
 eligible as traversal start nodes.

 If the indicated seed value is not present the traversal_node will not be chosen as a start traversal_node. This means that different seed values
 may generate different traversal orderings.

 For example,
   Dataset(
            name="mysql",
            after="mysql2"
            collections=[ Collection( name="users", after=("mysql","addresses")
                fields=[
                    Field(name="id"),
                    Field(name="email", identity= "email",
                    Field(name="name", references=[(("mongo", "users", "username")),"to")]),
        ],
    )

   means
 - create a traversal_node ("mysql", "users")
 - this traversal_node will not run until all nodes in the data dataset "mysql2" have run
 - this nodes "name" field references "mongo.users.username".
 - this traversal_node will starting with the seed value {"email":"test@test.com", "id":1} will be populated with
 email == test@test.com
 - this traversal_node will not be run until after ("mysql", "addresses")

 a traversal beginning with this traversal_node would then look like
  - select all values in mysql.users where email == "test@test.com"
  - select all values in mongo.users where the username is in (any of the "name" fields in the rows found in the previous step.

  Note that this makes no statements about _how_ this data is to be retrieved. In particular many data stores will need special
  instructions (such as json paths) to extract data.


"""

from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from re import match, search
from typing import Any, Callable, Dict, List, Literal, Optional, Set, Tuple, Union

from fideslang.models import FieldMaskingStrategyOverride, MaskingStrategyOverride
from fideslang.validation import FidesKey
from pydantic import BaseModel, ConfigDict, field_serializer, field_validator

from fides.api.common_exceptions import FidesopsException
from fides.api.graph.data_type import (
    DataType,
    DataTypeConverter,
    get_data_type_converter,
)
from fides.api.util.collection_util import merge_dicts
from fides.api.util.querytoken import QueryToken

DatasetAddress = str
SeedAddress = str
EdgeDirection = Literal["from", "to"]


class CollectionAddress:
    """The representation of a collection in the graph, specified by
    (data dataset name, collection name)"""

    def __init__(self, dataset: str, collection: str):
        self.dataset = dataset
        self.collection = collection
        self.value: str = ":".join((dataset, collection))

    def __repr__(self) -> str:
        return self.value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CollectionAddress):
            return False
        return other.value == self.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __lt__(self, other: CollectionAddress) -> bool:
        return self.value < other.value

    @staticmethod
    def from_string(address_str: str) -> CollectionAddress:
        """Generate a collection address from a string of the form 'dataset:collection'
        (that is, the form generated by Collection.__repr__)"""
        try:
            return CollectionAddress(*address_str.split(":"))
        except Exception:
            raise FidesopsException(
                f"'{address_str}' is not a valid collection address"
            )

    def field_address(self, field_path: FieldPath) -> FieldAddress:
        """Create a field address appended to this collection address.

        collection_address.field_address(FieldPath('a', 'b', 'c', 'd')) = dataset_name:collection_name:a.b.c.d
        """
        return FieldAddress(self.dataset, self.collection, *field_path.levels)


ROOT_COLLECTION_ADDRESS: CollectionAddress = CollectionAddress("__ROOT__", "__ROOT__")
"""Define a start traversal_node that all seed nodes begin from"""
TERMINATOR_ADDRESS = CollectionAddress("__TERMINATE__", "__TERMINATE__")
"""An address that corresponds to traversal termination"""


class FieldPath:
    """Fields are addressable by a (possibly) nested name. This key
    represents a field name held as a tuple of possibly descending levels.
    A scalar field is represented as a single-element tuple.

    Examples:
    FieldPath('a', 'b', 'c', 'd').levels = ('a', 'b', 'c', 'd')
    FieldPath('a', 'b', 'c', 'd').string_path = 'a.b.c.d'

    FieldPath('a').levels = ('a',)
    FieldPath('a').string_path = 'a'
    """

    def __init__(self, *names: str):
        self.levels: Tuple[str, ...] = tuple(names)
        self.string_path: str = ".".join(self.levels)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FieldPath):
            return False
        return other.levels == self.levels

    def __hash__(self) -> int:
        return hash(self.string_path)

    def __repr__(self) -> str:
        return f"FieldPath{self.levels}"

    def __lt__(self, other: "FieldPath") -> bool:
        return self.string_path < other.string_path

    def prepend(self, prefix: str) -> "FieldPath":
        """Return a new FieldPath with the prefix prepended."""
        return FieldPath(*((prefix,) + self.levels))

    @staticmethod
    def parse(path_str: str) -> FieldPath:
        """Create a FieldPath from a dot-separated input string"""
        return FieldPath(*path_str.split("."))


class FieldAddress:
    """The representation of a field location in the graph, specified by
    (data dataset name, collection name, field name, subfield name, ... )

    All values after the second are grouped to provide a FieldPath object.
    Additional values are understood to refer to nested field values.
    e.g. ("dataset", "collection", "a", "b", "c") creates a reference to
    dataset:collection:a.b.c
    """

    def __init__(self, dataset: str, collection: str, *fields: str):
        self.dataset = dataset
        self.collection = collection
        self.field_path: FieldPath = FieldPath(*fields)
        self.value: str = ":".join((dataset, collection, self.field_path.string_path))

    def is_member_of(self, collection_address: CollectionAddress) -> bool:
        """True if this field represents a field in the given collection address."""
        return (
            self.dataset == collection_address.dataset
            and self.collection == collection_address.collection
        )

    def collection_address(self) -> CollectionAddress:
        """Return the collection prefix of this field address."""
        return CollectionAddress(self.dataset, self.collection)

    @staticmethod
    def from_string(field_address_string: str) -> FieldAddress:
        """Creates a Field Address from a string - especially useful for instantiating
        Fields in Collections that are built from data in RequestTask.collection"""
        try:
            split_string = field_address_string.split(":")
            dataset = split_string[0]
            collection = split_string[1]
            fields = split_string[2]
            split_fields = fields.split(".")
            return FieldAddress(dataset, collection, *split_fields)
        except Exception:
            raise FidesopsException(
                f"'{field_address_string}' is not a valid field address"
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FieldAddress):
            return False
        return other.value == self.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return self.value

    def __lt__(self, other: FieldAddress) -> bool:
        return self.value < other.value


class Field(BaseModel, ABC):
    """A single piece of data"""

    name: str
    primary_key: bool = False
    references: List[Tuple[FieldAddress, Optional[EdgeDirection]]] = []
    """references to other fields in any other datasets"""
    identity: Optional[SeedAddress] = None
    """an optional pointer to an arbitrary key in an expected json package provided as a seed value"""
    data_categories: Optional[List[FidesKey]] = None
    data_type_converter: DataTypeConverter = DataType.no_op.value
    return_all_elements: Optional[bool] = None
    masking_strategy_override: Optional[FieldMaskingStrategyOverride] = None
    # Should field be returned by query if it is in an entrypoint array field, or just if it matches query?
    custom_request_field: Optional[str] = None

    """Known type of held data"""
    length: Optional[int] = None
    """Known length of held data"""

    is_array: bool = False

    read_only: Optional[bool] = None
    """Optionally specify if a field is read-only, meaning it can't be updated or deleted. """
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("data_type_converter")
    def serialize_data_type_converter(
        self, data_type_converter: DataTypeConverter
    ) -> Optional[str]:
        """Help Pydantic V2 serialize this unknown type"""
        return data_type_converter.name if data_type_converter.name else None

    @field_serializer("references")
    def serialize_references(
        self, references: List[Tuple[FieldAddress, Optional[EdgeDirection]]]
    ) -> List[Tuple[str, Optional[str]]]:
        """Help Pydantic V2 serialize this unknown type"""
        return [(ref[0].value, ref[1] if ref else None) for ref in references]

    @abstractmethod
    def cast(self, value: Any) -> Optional[Any]:
        """Cast the input value into the form represented by data_type."""

    def data_type(self) -> str:
        """return the data type name"""
        return self.data_type_converter.name

    @abstractmethod
    def collect_matching(self, func: Callable[[Field], bool]) -> Dict[FieldPath, Field]:
        """Find fields or subfields satisfying the input function"""

    def __repr__(self) -> str:
        """Overrides print method to be more succinct"""
        return f"{self.__class__.__name__}(name='{self.name}', data_type='{self.data_type()}', is_array={self.is_array})"


class ScalarField(Field):
    """A field that represents a simple value. Most fields will be scalar fields."""

    def cast(self, value: Any) -> Optional[Any]:
        """Cast the input value into the form represented by data_type.

        - If the data_type is None, then it has not been specified, so just return the input value.
        - Return either a cast value or None"""
        if not isinstance(value, QueryToken):
            return self.data_type_converter.to_value(value)

        return value

    def collect_matching(self, func: Callable[[Field], bool]) -> Dict[FieldPath, Field]:
        """Returns the field if it satisfies the input function"""
        if func(self):
            return {FieldPath(self.name): self}  # pylint: disable=no-member
        return {}

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ScalarField):
            return False

        return self.__dict__ == other.__dict__


class ObjectField(Field):
    """A field that represents a json dict structure."""

    fields: Dict[str, Field]

    def cast(self, value: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """Cast the input value into the form represented by data_type."""

        return {
            field.name: field.cast(value[field.name])  # pylint: disable=no-member
            for field in self.fields.values()
            if field.name in value
        }

    def collect_matching(self, func: Callable[[Field], bool]) -> Dict[FieldPath, Field]:
        """Find fields or subfields satisfying the input function

        Object fields will continue to call collect_matching until we get to the base case,
        which is a ScalarField.
        """
        base = (
            {FieldPath(self.name): self}  # pylint: disable=no-member
            if func(self)
            else {}
        )
        child_dicts = merge_dicts(
            *[field.collect_matching(func) for field in self.fields.values()]
        )
        return merge_dicts(
            base,  # type: ignore
            {
                field_path.prepend(self.name): field  # pylint: disable=no-member
                for field_path, field in child_dicts.items()
            },
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ObjectField):
            return False

        print(f"{self.__dict__=}")
        print(f"{other.__dict__=}")

        return self.__dict__ == other.__dict__


# pylint: disable=too-many-arguments
def generate_field(
    name: str,
    data_categories: Optional[List[FidesKey]],
    identity: Optional[str],
    data_type_name: str,
    references: List[Tuple[FieldAddress, Optional[EdgeDirection]]],
    is_pk: bool,
    length: Optional[int],
    is_array: bool,
    sub_fields: List[Field],
    return_all_elements: Optional[bool],
    read_only: Optional[bool],
    custom_request_field: Optional[str],
    masking_strategy_override: Optional[FieldMaskingStrategyOverride],
) -> Field:
    """Generate a graph field."""

    if sub_fields:
        return ObjectField(
            name=name,
            data_categories=data_categories,
            is_array=is_array,
            fields={f.name: f for f in sub_fields},
            data_type_converter=DataType.object.value,
            return_all_elements=return_all_elements,
        )
    return ScalarField(
        name=name,
        data_categories=data_categories,
        identity=identity,
        data_type_converter=get_data_type_converter(data_type_name),
        references=references,
        primary_key=is_pk,
        length=length,
        is_array=is_array,
        return_all_elements=return_all_elements,
        read_only=read_only,
        custom_request_field=custom_request_field,
        masking_strategy_override=masking_strategy_override,
    )


@dataclass
class MaskingTruncation:
    """Data class to store truncation params related to data masking"""

    data_type_converter: Optional[DataTypeConverter]
    length: Optional[int]


# for now, we only support BQ partitioning, so the clause pattern we expect is BQ-specific
BIGQUERY_PARTITION_CLAUSE_PATTERN = r"^`(?P<field_1>[a-zA-Z0-9_]*)` ([<|>][=]?) (?P<operand_1>[a-zA-Z0-9_\s(),\.\"\'\-]*)(\sAND `(?P<field_2>[a-zA-Z0-9_]*)` ([<|>][=]?) (?P<operand_2>[a-zA-Z0-9_\s(),\.\"\'\-]*))?$"
# protected keywords that are _not_ allowed in the operands, to avoid any potential malicious execution.
PROHIBITED_KEYWORDS = [
    "UNION",
    "INSERT",
    "UPDATE",
    "CREATE",
    "DROP",
    "SELECT",
    "CHAR",
    "HAVING",
    "EXEC",
]


class Collection(BaseModel):
    """A single grouping of individual data points that are accessed together"""

    name: str
    skip_processing: Optional[bool] = False
    fields: List[Field]
    # an optional list of collections that this collection must run after
    after: Set[CollectionAddress] = set()
    erase_after: Set[CollectionAddress] = set()
    # An optional set of dependent fields that need to be queried together
    grouped_inputs: Set[str] = set()
    data_categories: Set[FidesKey] = set()
    masking_strategy_override: Optional[MaskingStrategyOverride] = None
    partitioning: Optional[Dict] = None

    @property
    def field_dict(self) -> Dict[FieldPath, Field]:
        """Maps FieldPaths to Fields

        Flattens all the Fields so they are on one level: all nested fields are brought to the top.
        """
        return self.recursively_collect_matches(lambda f: True)

    @property
    def top_level_field_dict(self) -> Dict[FieldPath, Field]:
        """Returns a map of top-level FieldPaths mapped to fields"""
        return {FieldPath(field.name): field for field in self.fields}

    def recursively_collect_matches(
        self, func: Callable[[Field], bool]
    ) -> Dict[FieldPath, Field]:
        """Recurse through fields and subfields, creating a flattened dictionary
        of field paths mapped to fields where the function is satisfied"""
        matches = [field.collect_matching(func) for field in self.fields]
        return merge_dicts(*matches)

    def references(
        self,
    ) -> Dict[FieldPath, List[Tuple[FieldAddress, Optional[EdgeDirection]]]]:
        """return references from fields in this collection to fields in any other collection

        A nested field can be a reference.
        """
        return {
            field_path: field.references
            for field_path, field in self.field_dict.items()
            if field.references
        }

    def identities(self) -> Dict[FieldPath, str]:
        """return identity pointers included in the table"""
        return {
            field_path: field.identity  # type: ignore
            for field_path, field in self.field_dict.items()
            if field.identity
        }

    def custom_request_fields(self) -> Dict[FieldPath, str]:
        """
        Return custom request fields included in the table,
        i.e fields whose values may come in a custom request field on a DSR.

        E.g if the collection is defined like:
        - name: publishers
        - fields:
            - name: id
              fides_meta:
                identity: true
            - name: site_id
              fides_meta:
                custom_request_field: tenant_id

        Then this returns a dictionary of the form {FieldPath("site_id"): "tenant_id"}
        """
        return {
            field_path: field.custom_request_field
            for field_path, field in self.field_dict.items()
            if field.custom_request_field
        }

    def field(self, field_path: FieldPath) -> Optional[Field]:
        """Return Field (looked up by FieldPath) if on Collection or None if not found"""
        return self.field_dict[field_path] if field_path in self.field_dict else None

    @property
    def field_paths_by_category(self) -> Dict[FidesKey, List[FieldPath]]:
        """Returns mapping of data categories to a list of FieldPaths, flips FieldPaths -> categories
        to be categories -> FieldPaths.

        Example:
            {
                "user.contact.address.city": [FieldPath("city")],
                "user.contact.address.street": [FieldPath("house"), FieldPath("street")],
                "system.operations": ["id"],
                "user.contact.address.state": [FieldPath("state", "code"),FieldPath("state", "full_name"), ],
                "user.contact.address.postal_code": ["zip"]
            }
        """
        categories = defaultdict(list)
        for field_path, field in self.field_dict.items():
            for category in field.data_categories or []:
                categories[category].append(field_path)
        return categories

    def contains_field(self, func: Callable[[Field], bool]) -> bool:
        """True if any field in this collection matches the condition of the callable

        Currently used to assert at least one field in the collection contains a primary
        key before erasing
        """
        return any(self.recursively_collect_matches(func))

    @classmethod
    def parse_from_request_task(cls, data: Dict) -> Collection:
        """
        Take raw collection data saved on RequestTask.collection and converts it back into a Collection.

        See Config > json_encoders for some of the fields that needed special handling for serialization for
        database storage.
        """
        data = copy.deepcopy(data)

        def build_field(serialized_field: dict) -> Field:
            """Convert a serialized field on RequestTask.collection.fields into a Scalar Field
            or Object Field"""
            converted_references: List[Tuple[FieldAddress, Optional[EdgeDirection]]] = (
                []
            )
            for reference in serialized_field.pop("references", []):
                field_address: str = reference[0]
                edge_direction: Optional[str] = reference[1]
                converted_references.append(
                    (FieldAddress.from_string(field_address), edge_direction)  # type: ignore
                )

            data_type_converter: DataTypeConverter = get_data_type_converter(
                serialized_field.pop("data_type_converter")
            )

            # We can't convert the fields to abstract class Field - they need to be proper
            # Scalar or ObjectFields
            converted: Union[ObjectField, ScalarField]
            if serialized_field.get("fields"):
                # Recursively build nested fields under Object field
                serialized_field["fields"] = {
                    field_name: build_field(fld)
                    for field_name, fld in serialized_field["fields"].items()
                }
                converted = ObjectField.model_validate(serialized_field)
                converted.references = converted_references
                converted.data_type_converter = data_type_converter
                return converted

            converted = ScalarField.model_validate(serialized_field)
            converted.references = converted_references
            converted.data_type_converter = data_type_converter
            return converted

        converted_fields = []
        for field in data.pop("fields"):
            converted_fields.append(build_field(field))

        data["fields"] = converted_fields
        data["after"] = {
            CollectionAddress.from_string(addr_string)
            for addr_string in data.get("after", [])
        }
        data["erase_after"] = {
            CollectionAddress.from_string(addr_string)
            for addr_string in data.get("erase_after", [])
        }
        data["partitioning"] = data.get("partitioning")

        return Collection.model_validate(data)

    @field_serializer("data_type_converter", check_fields=False)
    def serialize_data_type_converter(
        self, data_type_converter: DataTypeConverter
    ) -> Optional[str]:
        """Help Pydantic V2 serialize this unknown type"""
        return data_type_converter.name if data_type_converter.name else None

    @field_serializer("after")
    def serialize_after(self, after: Set[CollectionAddress]) -> Set[str]:
        """Help Pydantic V2 serialize this unknown type"""
        return {aft.value for aft in after}

    @field_serializer("erase_after")
    def serialize_erase_after(self, erase_after: Set[CollectionAddress]) -> Set[str]:
        """Help Pydantic V2 serialize this unknown type"""
        return {aft.value for aft in erase_after}

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        # This supports running Collection.json() to serialize less standard
        # types so it can be saved to the database under RequestTask.collection
        json_encoders={
            Set: lambda val: list(  # pylint: disable=unhashable-member,unnecessary-lambda
                val
            ),
            DataTypeConverter: lambda dtc: dtc.name if dtc.name else None,
            FieldAddress: lambda fa: fa.value,
            CollectionAddress: lambda ca: ca.value,
        },
    )

    @field_validator("partitioning")
    @classmethod
    def validate_partitioning(
        cls, partitioning: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        """
        Validates the `partitioning` dict field.

        The `partitioning` dict field is untyped in Fideslang, but here we enforce
        that it has the required and expected `where_clauses` key, whose value must be
        a list of strings.

        The string values are validated to ensure they match the expected syntax, which
        is strictly prescribed. The string values MUST be a valid SQL clause that defines
        a partition window, with the form:

        ```
        `column_1` >(=) [some value] AND `column_1` <(=) [some value]
        ```

        To be clear, some notable constraints on the input:
        - the clause string must begin by referencing a column name wrapped by backticks (`)
        - the clause string must compare that first column with a `<>(=)` operator, and may
        include at most one other conditional with a `<>(=)` operator that's joined to the first
        conditional via an AND operator
        - if the clause string contains a second conditional, it must reference the same column name
        as the first conditional, also wrapped by backticks
        - column names (wrapped by backticks) must always be on the _left_ side of the `<>(=)`operator
        in its conditional

        """
        if not partitioning:
            return partitioning

        # NOTE: when we deprecate `where_clause` partitioning in favor of a more proper partitioning DSL,
        # we should be sure to still support the existing `where_clause` partition definition on
        # any in-progress DSRs so that they can run through to completion.
        if where_clauses := partitioning.get("where_clauses"):
            if not isinstance(where_clauses, List) or not all(
                isinstance(where_clause, str) for where_clause in where_clauses
            ):
                raise ValueError("`where_clauses` must be a list of strings!")
            for partition_clause in where_clauses:
                if matching := match(
                    BIGQUERY_PARTITION_CLAUSE_PATTERN, partition_clause
                ):
                    # check that if there are two field comparison sub-clauses, they reference the same field, e.g.:
                    # "`my_field_1` > 5 AND `my_field_1` <= 10", not "`my_field_1` > 5 AND `my_field_1` <= 10"
                    if matching["field_2"] is not None and (
                        matching["field_1"] != matching["field_2"]
                    ):
                        raise ValueError(
                            f"Partition clause must have matching fields. Identified non-matching field references '{matching['field_1']}' and '{matching['field_2']}"
                        )

                    for prohibited_keyword in PROHIBITED_KEYWORDS:
                        search_str = prohibited_keyword.lower() + r"\s"
                        if search(search_str, partition_clause.lower()):
                            raise ValueError(
                                "Prohibited keyword referenced in partition clause"
                            )
                else:
                    raise ValueError("Unsupported partition clause format")
            return partitioning
        raise ValueError(
            "`where_clauses` must be specified in `partitioning` specification!"
        )


class GraphDataset(BaseModel):
    """Master collection of collections that are accessed in a common way"""

    name: str
    collections: List[Collection]
    # an optional list of datasets that this dataset must run after
    after: Set[DatasetAddress] = set()
    # ConnectionConfig key
    connection_key: FidesKey



"""property specific messaging db models

Revision ID: 2736c942faa2
Revises: 52a5f1a957bc
Create Date: 2024-05-28 14:26:09.012859

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2736c942faa2"
down_revision = "52a5f1a957bc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "messaging_template_to_property",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("messaging_template_id", sa.String(), nullable=False),
        sa.Column("property_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["messaging_template_id"],
            ["messaging_template.id"],
        ),
        sa.ForeignKeyConstraint(
            ["property_id"],
            ["plus_property.id"],
        ),
        sa.PrimaryKeyConstraint("id", "messaging_template_id", "property_id"),
        sa.UniqueConstraint(
            "messaging_template_id",
            "property_id",
            name="messaging_template_id_property_id",
        ),
    )
    op.create_index(
        op.f("ix_messaging_template_to_property_id"),
        "messaging_template_to_property",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_messaging_template_to_property_messaging_template_id"),
        "messaging_template_to_property",
        ["messaging_template_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_messaging_template_to_property_property_id"),
        "messaging_template_to_property",
        ["property_id"],
        unique=False,
    )
    op.add_column(
        "messaging_template", sa.Column("is_enabled", sa.Boolean(), nullable=True)
    )
    op.execute("UPDATE messaging_template SET is_enabled = FALSE;")

    op.alter_column("messaging_template", "is_enabled", nullable=False)

    op.alter_column("messaging_template", "key", new_column_name="type")

    op.drop_index(op.f("ix_messaging_template_key"), table_name="messaging_template")
    op.create_index(
        op.f("ix_messaging_template_type"), "messaging_template", ["type"], unique=False
    )
    op.add_column(
        "plus_property",
        sa.Column("is_default", sa.Boolean(), server_default="f", nullable=True),
    )
    op.create_index(
        "only_one_default",
        "plus_property",
        ["is_default"],
        unique=True,
        postgresql_where=sa.text("is_default"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "only_one_default",
        table_name="plus_property",
        postgresql_where=sa.text("is_default"),
    )
    op.drop_column("plus_property", "is_default")
    op.drop_index(op.f("ix_messaging_template_type"), table_name="messaging_template")
    op.alter_column("messaging_template", "type", new_column_name="key")
    op.create_index(
        "ix_messaging_template_key", "messaging_template", ["key"], unique=False
    )
    op.drop_column("messaging_template", "is_enabled")
    op.drop_index(
        op.f("ix_messaging_template_to_property_property_id"),
        table_name="messaging_template_to_property",
    )
    op.drop_index(
        op.f("ix_messaging_template_to_property_messaging_template_id"),
        table_name="messaging_template_to_property",
    )
    op.drop_index(
        op.f("ix_messaging_template_to_property_id"),
        table_name="messaging_template_to_property",
    )
    op.drop_table("messaging_template_to_property")
    # ### end Alembic commands ###



from typing import List

import pg8000
from google.cloud.sql.connector import Connector
from google.oauth2 import service_account
from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import (  # type: ignore
    Connection,
    Engine,
    LegacyCursorResult,
    create_engine,
)

from fides.api.graph.execution import ExecutionNode
from fides.api.schemas.connection_configuration.connection_secrets_google_cloud_sql_postgres import (
    GoogleCloudSQLPostgresSchema,
)
from fides.api.service.connectors.query_configs.google_cloud_postgres_query_config import (
    GoogleCloudSQLPostgresQueryConfig,
)
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row
from fides.config import get_config

CONFIG = get_config()


class GoogleCloudSQLPostgresConnector(SQLConnector):
    """Connector specific to Google Cloud SQL for Postgres"""

    secrets_schema = GoogleCloudSQLPostgresSchema

    @property
    def default_db_name(self) -> str:
        """Default database name for Google Cloud SQL Postgres"""
        return "postgres"

    # Overrides SQLConnector.create_client
    def create_client(self) -> Engine:
        """Returns a SQLAlchemy Engine that can be used to interact with a database"""

        config = self.secrets_schema(**self.configuration.secrets or {})

        credentials = service_account.Credentials.from_service_account_info(
            dict(config.keyfile_creds)
        )

        # initialize connector with the loaded credentials
        connector = Connector(credentials=credentials)

        def getconn() -> pg8000.dbapi.Connection:
            conn: pg8000.dbapi.Connection = connector.connect(
                config.instance_connection_name,
                "pg8000",
                user=config.db_iam_user,
                db=config.dbname or self.default_db_name,
                enable_iam_auth=True,
            )
            return conn

        return create_engine("postgresql+pg8000://", creator=getconn)

    @staticmethod
    def cursor_result_to_rows(results: LegacyCursorResult) -> List[Row]:
        """results to a list of dictionaries"""
        return SQLConnector.default_cursor_result_to_rows(results)

    def build_uri(self) -> None:
        """
        We need to override this method so it is not abstract anymore, and GoogleCloudSQLPostgresConnector is instantiable.
        """

    def set_schema(self, connection: Connection) -> None:
        """Sets the schema for a postgres database if applicable"""
        config = self.secrets_schema(**self.configuration.secrets or {})
        if config.db_schema:
            logger.info("Setting PostgreSQL search_path before retrieving data")
            stmt = text("SELECT set_config('search_path', :search_path, false)")
            stmt = stmt.bindparams(search_path=config.db_schema)
            connection.execute(stmt)

    # Overrides SQLConnector.query_config
    def query_config(self, node: ExecutionNode) -> GoogleCloudSQLPostgresQueryConfig:
        """Query wrapper corresponding to the input execution_node."""
        return GoogleCloudSQLPostgresQueryConfig(node)



"""update default timestamps for execution logs

Revision ID: c90d46f6d3f2
Revises: abc53a00755b
Create Date: 2024-11-15 17:29:29.934671

"""

from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "c90d46f6d3f2"
down_revision = "abc53a00755b"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        "executionlog", "created_at", server_default=text("clock_timestamp()")
    )

    op.alter_column(
        "executionlog", "updated_at", server_default=text("clock_timestamp()")
    )


def downgrade():
    op.alter_column("executionlog", "created_at", server_default=text("now()"))

    op.alter_column("executionlog", "updated_at", server_default=text("now()"))



from typing import TYPE_CHECKING, Annotated, List, Optional, Set

from fideslang.validation import FidesKey
from pydantic import ConfigDict, Field, StringConstraints, field_validator

from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.connection_configuration.connection_config import (
    ConnectionConfigurationResponse,
)
from fides.api.util.text import to_snake_case

DSRLabelFieldType = Annotated[
    str, StringConstraints(max_length=200, strip_whitespace=True)
]


class ManualWebhookField(FidesSchema):
    """Schema to describe the attributes on a manual webhook field"""

    pii_field: Annotated[
        str, StringConstraints(min_length=1, max_length=200, strip_whitespace=True)
    ]
    dsr_package_label: Optional[DSRLabelFieldType] = None
    data_categories: Optional[List[FidesKey]] = None

    @field_validator("dsr_package_label")
    @classmethod
    def convert_empty_string_dsr_package_label(
        cls, value: Optional[str]
    ) -> Optional[str]:
        """
        We specifically allow the dsr_package_label to be submitted as an empty string on input,
        so converting to None here.
        """
        return None if value == "" else value

    model_config = ConfigDict(from_attributes=True)


if TYPE_CHECKING:
    ManualWebhookFieldsList = List[ManualWebhookField]
else:
    ManualWebhookFieldsList = Annotated[List[ManualWebhookField], Field(min_length=1)]


class AccessManualWebhooks(FidesSchema):
    """Expected request body for creating Access Manual Webhooks"""

    fields: ManualWebhookFieldsList
    model_config = ConfigDict(from_attributes=True)

    @field_validator("fields")
    @classmethod
    def check_for_duplicates(
        cls, value: List[ManualWebhookField]
    ) -> List[ManualWebhookField]:
        """
        Verify that pii_fields and dsr_package_labels are unique.

        Set the dsr_package_label to a snake_cased lower case version of pii field if it doesn't exist.
        """
        unique_pii_fields: Set[str] = {field.pii_field for field in value}
        if len(value) != len(unique_pii_fields):
            raise ValueError("pii_fields must be unique")

        for field in value:
            if not field.dsr_package_label:
                field.dsr_package_label = DSRLabelFieldType(
                    to_snake_case(field.pii_field)
                )

        unique_dsr_package_labels: Set[Optional[str]] = {
            field.dsr_package_label for field in value
        }
        if len(value) != len(unique_dsr_package_labels):
            # Postponing dsr_package_label uniqueness check in case we get overlaps
            # above when we fallback to converting pii_fields to dsr_package_labels
            raise ValueError("dsr_package_labels must be unique")

        return value

    @field_validator("fields")
    @classmethod
    def fields_must_exist(
        cls, value: List[ManualWebhookField]
    ) -> List[ManualWebhookField]:
        """
        Verify that pii_fields and dsr_package_labels are unique.

        Set the dsr_package_label to a snake_cased lower case version of pii field if it doesn't exist.
        """
        unique_pii_fields: Set[str] = {field.pii_field for field in value}
        if len(value) != len(unique_pii_fields):
            raise ValueError("pii_fields must be unique")

        for field in value:
            if not field.dsr_package_label:
                field.dsr_package_label = DSRLabelFieldType(
                    to_snake_case(field.pii_field)
                )

        unique_dsr_package_labels: Set[Optional[str]] = {
            field.dsr_package_label for field in value
        }
        if len(value) != len(unique_dsr_package_labels):
            raise ValueError("dsr_package_labels must be unique")

        return value


class AccessManualWebhookResponse(AccessManualWebhooks):
    """Expected response for accessing Access Manual Webhooks"""

    connection_config: ConnectionConfigurationResponse
    id: str
    model_config = ConfigDict(from_attributes=True)



"""add property id to historical records

Revision ID: 57479150efd6
Revises: fc2b2c06e595
Create Date: 2024-05-14 18:26:57.646508

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "57479150efd6"
down_revision = "ad0109b041b3"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacypreferencehistory", sa.Column("property_id", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("property_id", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("servednoticehistory", "property_id")
    op.drop_column("privacypreferencehistory", "property_id")



from typing import TYPE_CHECKING, Optional, Protocol

from fides.api.models.policy import Policy

if TYPE_CHECKING:
    from fides.api.graph.traversal import TraversalNode


class NodeFilter(Protocol):
    """Protocol for node filters that determine if nodes should be excluded from traversal"""

    def exclude_node(self, node: "TraversalNode") -> bool:
        """Returns True if the node should be excluded from traversal requirements"""


class CustomRequestFieldFilter:
    """
    Remove nodes that have custom request fields, since we don't care if these are reachable or not.
    They will be used independently by the Dynamic Email Erasure Connector.

    TODO: ideally we'll update the Traversal code to include these "custom request field datasets"
    as part of the main graph. This is a targeted workaround for now.
    """

    def exclude_node(self, node: "TraversalNode") -> bool:
        return bool(node.node.collection.custom_request_fields())


class PolicyDataCategoryFilter:
    """
    If a policy is provided, we will allow collections without relevant
    data categories to remain unreachable without raising any exceptions.
    """

    def __init__(self, policy: Optional[Policy]):
        self.policy = policy

    def exclude_node(self, node: "TraversalNode") -> bool:
        if not self.policy:
            return False
        return not self.policy.applies_to(node)



# pylint: disable=too-many-lines,too-many-statements
import copy
import traceback
from abc import ABC
from functools import wraps
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from loguru import logger
from ordered_set import OrderedSet
from sqlalchemy.orm import Session

from fides.api.api.deps import get_db_contextmanager as get_db
from fides.api.common_exceptions import (
    ActionDisabled,
    AwaitingAsyncTaskCallback,
    CollectionDisabled,
    NotSupportedForCollection,
    PrivacyRequestErasureEmailSendRequired,
    SkippingConsentPropagation,
)
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    CollectionAddress,
    Field,
    FieldAddress,
    FieldPath,
    GraphDataset,
)
from fides.api.graph.execution import ExecutionNode
from fides.api.graph.graph import DatasetGraph
from fides.api.graph.traversal import Traversal, TraversalNode
from fides.api.models.connectionconfig import (
    AccessLevel,
    ConnectionConfig,
    ConnectionType,
)
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import ExecutionLog, PrivacyRequest, RequestTask
from fides.api.schemas.policy import ActionType, CurrentStep
from fides.api.schemas.privacy_request import ExecutionLogStatus
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.task.consolidate_query_matches import consolidate_query_matches
from fides.api.task.filter_element_match import filter_element_match
from fides.api.task.refine_target_path import FieldPathNodeInput
from fides.api.task.task_resources import TaskResources
from fides.api.util.cache import get_cache
from fides.api.util.collection_util import (
    NodeInput,
    Row,
    append_unique,
    extract_key_for_address,
    make_immutable,
    make_mutable,
)
from fides.api.util.consent_util import add_errored_system_status_for_consent_reporting
from fides.api.util.logger import Pii
from fides.api.util.logger_context_utils import LoggerContextKeys
from fides.api.util.saas_util import FIDESOPS_GROUPED_INPUTS
from fides.config import CONFIG

COLLECTION_FIELD_PATH_MAP = Dict[CollectionAddress, List[Tuple[FieldPath, FieldPath]]]

EMPTY_REQUEST = PrivacyRequest()
EMPTY_REQUEST_TASK = RequestTask()


def retry(
    action_type: ActionType,
    default_return: Any,
) -> Callable:
    """
    Retry decorator for access and right to forget requests requests -

    If an exception is raised, we retry the function `count` times with exponential backoff. After the number of
    retries have expired, we call GraphTask.end() with the appropriate `action_type` and `default_return`.

    If we exceed the number of TASK_RETRY_COUNT retries, we re-raise the exception to stop execution of the privacy request.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def result(*args: Any, **kwargs: Any) -> Any:
            func_delay = CONFIG.execution.task_retry_delay
            method_name = func.__name__
            self = args[0]

            raised_ex: Optional[Union[BaseException, Exception]] = None
            for attempt in range(CONFIG.execution.task_retry_count + 1):
                try:
                    self.skip_if_disabled()
                    self.skip_if_action_disabled(action_type)
                    # Create ExecutionLog with status in_processing or retrying
                    if attempt:
                        self.log_retry(action_type)
                    else:
                        self.log_start(action_type)
                    # Run access or erasure request
                    return func(*args, **kwargs)
                except AwaitingAsyncTaskCallback as ex:
                    traceback.print_exc()
                    logger.warning(
                        "Request Task {} {} {} awaiting async callback",
                        self.request_task.id if self.request_task.id else None,
                        method_name,
                        self.execution_node.address,
                    )
                    self.log_awaiting_processing(action_type, ex)
                    # Request Task put in "awaiting_processing" status and exited, awaiting Async Callback
                    return None
                except PrivacyRequestErasureEmailSendRequired as exc:
                    traceback.print_exc()
                    self.request_task.rows_masked = 0
                    self.log_end(action_type, ex=None, success_override_msg=exc)
                    self.resources.cache_erasure(
                        f"{self.traversal_node.address.value}", 0
                    )  # Cache that the erasure was performed in case we need to restart for DSR 2.0
                    return 0
                except (
                    CollectionDisabled,
                    ActionDisabled,
                    NotSupportedForCollection,
                ) as exc:
                    traceback.print_exc()
                    logger.warning(
                        "Skipping collection {} for privacy_request: {}",
                        self.execution_node.address,
                        self.resources.request.id,
                    )
                    self.log_skipped(action_type, exc)
                    return default_return
                except SkippingConsentPropagation as exc:
                    traceback.print_exc()
                    logger.warning(
                        "Skipping consent propagation on collection {} for privacy_request: {}",
                        self.execution_node.address,
                        self.resources.request.id,
                    )
                    self.log_skipped(action_type, exc)
                    for pref in self.resources.request.privacy_preferences:
                        # For consent reporting, also caching the given system as skipped for all historical privacy preferences.
                        pref.cache_system_status(
                            self.resources.session,
                            self.connector.configuration.system_key,
                            ExecutionLogStatus.skipped,
                        )
                    return default_return
                except BaseException as ex:  # pylint: disable=W0703
                    traceback.print_exc()
                    func_delay *= CONFIG.execution.task_retry_backoff
                    logger.warning(
                        "Retrying {} {} in {} seconds...",
                        method_name,
                        self.execution_node.address,
                        func_delay,
                    )
                    sleep(func_delay)
                    raised_ex = ex
            self.log_end(action_type, raised_ex)
            # transform ActionType -> CurrentStep type, expected by cache_failed_checkpoint_details
            self.resources.request.cache_failed_checkpoint_details(
                step=CurrentStep[
                    action_type.value
                ]  # Convert ActionType into a CurrentStep, no longer coerced with Pydantic V2
            )
            add_errored_system_status_for_consent_reporting(
                self.resources.session,
                self.resources.request,
                self.connector.configuration,
            )
            if not self.request_task.id:
                # TODO Remove when we stop support for DSR 2.0
                # Re-raise to stop privacy request execution on failure for
                # deprecated DSR 2.0 sequential execution
                raise raised_ex  # type: ignore
            return default_return

        return result

    return decorator


def mark_current_and_downstream_nodes_as_failed(
    privacy_request_task: RequestTask, db: Session
) -> None:
    """
    For DSR 3.0, if the current node fails, mark it and *every descendant that can be reached by the current node*
    as failed
    """
    if not privacy_request_task.id:
        return

    logger.info(f"Marking task {privacy_request_task.id} and descendants as errored")

    privacy_request_task.status = ExecutionLogStatus.error
    db.add(privacy_request_task)

    for descendant_addr in privacy_request_task.all_descendant_tasks or []:
        descendant: Optional[RequestTask] = (
            privacy_request_task.get_tasks_with_same_action_type(db, descendant_addr)
            .filter(RequestTask.status == ExecutionLogStatus.pending)
            .first()
        )
        if not descendant:
            continue
        descendant.status = ExecutionLogStatus.error
        db.add(descendant)

    db.commit()


class GraphTask(ABC):  # pylint: disable=too-many-instance-attributes
    """A task that operates on one traversal_node of a traversal"""

    def __init__(
        self, resources: TaskResources
    ):  # cache config, log config, db store config
        super().__init__()
        self.request_task = resources.privacy_request_task
        self.execution_node = ExecutionNode(resources.privacy_request_task)
        self.resources = resources
        self.connector: BaseConnector = resources.get_connector(
            self.execution_node.connection_key  # ConnectionConfig.key
        )

        self.key: CollectionAddress = self.execution_node.address

        self.execution_log_id = None
        # a local copy of the execution log record written to. If we write multiple status
        # updates, we will use this id to ensure that we're updating rather than creating
        # a new record

    def __repr__(self) -> str:
        return f"{type(self)}:{self.key}"

    def generate_dry_run_query(self) -> Optional[str]:
        """Type-specific query generated for this traversal_node."""
        return self.connector.dry_run_query(self.execution_node)

    def can_write_data(self) -> bool:
        """Checks if the relevant ConnectionConfig has been granted "write" access to its data"""
        connection_config: ConnectionConfig = self.connector.configuration
        return connection_config.access == AccessLevel.write

    def _combine_seed_data(
        self,
        *data: List[Row],
        grouped_data: Dict[str, Any],
        dependent_field_mappings: COLLECTION_FIELD_PATH_MAP,
    ) -> Dict[str, Any]:
        """Combine the seed data with the other dependent inputs. This is used when the seed data in a collection requires
        inputs from another collection to generate subsequent queries."""
        # Get the identity values from the seeds that were passed into this collection.
        seed_index = self.execution_node.input_keys.index(ROOT_COLLECTION_ADDRESS)
        seed_data = data[seed_index]

        for foreign_field_path, local_field_path in dependent_field_mappings[
            ROOT_COLLECTION_ADDRESS
        ]:
            dependent_values = consolidate_query_matches(
                row=seed_data, target_path=foreign_field_path  # type: ignore
            )
            grouped_data[local_field_path.string_path] = dependent_values
        return grouped_data

    def pre_process_input_data(
        self, *data: List[Row], group_dependent_fields: bool = False
    ) -> NodeInput:
        """
        Consolidates the outputs of queries from potentially multiple collections whose
        data is needed as input into the current collection.

        Each dict in the input list represents the output of a dependent task.
        These outputs should correspond to the input key order.  Any nested fields are
        converted into dot-separated paths in the return.

         table1: [{x:1, y:A}, {x:2, y:B}], table2: [{x:3},{x:4}], table3: [{z: {a: C}, "y": [4, 5]}]
           where table1.x => self.id,
           table1.y => self.name,
           table2.x => self.id
           table3.z.a => self.contact.address
           table3.y => self.contact.email
         becomes
         {id:[1,2,3,4], name:["A","B"], contact.address:["C"], "contact.email": [4, 5]}

         If there are dependent fields from one collection into another, they are separated out as follows:
         {fidesops_grouped_inputs: [{"organization_id": 1, "project_id": "math}, {"organization_id": 5, "project_id": "science"}]

         The output dictionary is constructed with deduplicated values for each key, ensuring that the value lists
         and the fides_grouped_input list contain only unique elements.
        """
        if not len(data) == len(self.execution_node.input_keys):
            logger.warning(
                "{} expected {} input keys, received {}",
                self,
                len(self.execution_node.input_keys),
                len(data),
            )

        # the ordered set is just to have a consistent output for testing, the order is not needed otherwise
        output: Dict[str, OrderedSet] = {FIDESOPS_GROUPED_INPUTS: OrderedSet()}

        (
            independent_field_mappings,
            dependent_field_mappings,
        ) = self.execution_node.build_incoming_field_path_maps(group_dependent_fields)

        for i, rowset in enumerate(data):
            collection_address = self.execution_node.input_keys[i]

            if (
                group_dependent_fields
                and self.execution_node.dependent_identity_fields
                and collection_address == ROOT_COLLECTION_ADDRESS
            ):
                # Skip building data for the root collection if the seed data needs to be combined with other inputs
                continue

            logger.info(
                "Consolidating incoming data into {} from {}.",
                self.execution_node.address,
                collection_address,
            )
            for row in rowset:
                # Consolidate lists of independent field inputs
                for foreign_field_path, local_field_path in independent_field_mappings[
                    collection_address
                ]:
                    new_values: List = consolidate_query_matches(
                        row=row, target_path=foreign_field_path
                    )
                    if new_values:
                        append_unique(output, local_field_path.string_path, new_values)

                # Separately group together dependent inputs if applicable
                if dependent_field_mappings[collection_address]:
                    grouped_data: Dict[str, Any] = {}
                    for (
                        foreign_field_path,
                        local_field_path,
                    ) in dependent_field_mappings[collection_address]:
                        dependent_values: List = consolidate_query_matches(
                            row=row, target_path=foreign_field_path
                        )
                        grouped_data[local_field_path.string_path] = dependent_values

                    if self.execution_node.dependent_identity_fields:
                        grouped_data = self._combine_seed_data(
                            *data,
                            grouped_data=grouped_data,
                            dependent_field_mappings=dependent_field_mappings,
                        )

                    output[FIDESOPS_GROUPED_INPUTS].add(make_immutable(grouped_data))

        return make_mutable(output)

    def update_status(
        self,
        msg: str,
        fields_affected: Any,
        action_type: ActionType,
        status: ExecutionLogStatus,
    ) -> None:
        """Update status activities - create an execution log (which stores historical logs)
        and update the Request Task's current status.
        """
        with get_db() as db:
            ExecutionLog.create(
                db=db,
                data={
                    "connection_key": self.execution_node.connection_key,
                    "dataset_name": self.execution_node.address.dataset,
                    "collection_name": self.execution_node.address.collection,
                    "fields_affected": fields_affected,
                    "action_type": action_type,
                    "status": status,
                    "privacy_request_id": self.resources.request.id,
                    "message": msg,
                },
            )

            # For DSR 3.0, updating the Request Task status when the ExecutionLog is
            # created to keep these in sync.
            # TODO remove conditional above alongside deprecating DSR 2.0
            if self.request_task.id:
                # Merge the request task to ensure we keep the changes made
                # to self.request_task that haven't yet been committed to the database
                request_task = db.merge(self.request_task)
                request_task.update_status(db, status)
                self.request_task = request_task

    def log_start(self, action_type: ActionType) -> None:
        """Task start activities"""
        logger.info("Starting node {}", self.key)

        self.update_status(
            "starting", [], action_type, ExecutionLogStatus.in_processing
        )

    def log_retry(self, action_type: ActionType) -> None:
        """Task retry activities"""
        logger.info("Retrying node {}", self.key)

        self.update_status("retrying", [], action_type, ExecutionLogStatus.retrying)

    def log_awaiting_processing(
        self, action_type: ActionType, ex: Optional[BaseException]
    ) -> None:
        """On paused activities"""
        logger.info("Pausing node {}", self.key)

        self.update_status(
            str(ex), [], action_type, ExecutionLogStatus.awaiting_processing
        )

    def log_skipped(self, action_type: ActionType, ex: str) -> None:
        """Log that a collection was skipped.  For now, this is because a collection has been disabled."""
        logger.info("Skipping node {}", self.key)
        if action_type == ActionType.consent and self.request_task.id:
            self.request_task.consent_sent = False
        self.update_status(str(ex), [], action_type, ExecutionLogStatus.skipped)

    def log_end(
        self,
        action_type: ActionType,
        ex: Optional[BaseException] = None,
        success_override_msg: Optional[BaseException] = None,
    ) -> None:
        """On completion activities"""
        if ex:
            logger.warning(
                "Ending {}, {} with failure {}",
                self.resources.request.id,
                self.key,
                Pii(ex),
            )
            self.update_status(str(ex), [], action_type, ExecutionLogStatus.error)
            # For DSR 3.0, Hooking into the GraphTask.log_end method to also mark the current
            # Request Task and every Request Task that can be reached from the current
            # task as errored.

            with get_db() as db:
                request_task = db.merge(self.request_task)
                mark_current_and_downstream_nodes_as_failed(request_task, db)
        else:
            logger.info("Ending {}, {}", self.resources.request.id, self.key)
            self.update_status(
                str(success_override_msg) if success_override_msg else "success",
                build_affected_field_logs(
                    self.execution_node, self.resources.policy, action_type
                ),
                action_type,
                ExecutionLogStatus.complete,
            )

    def post_process_input_data(
        self, pre_processed_inputs: NodeInput
    ) -> FieldPathNodeInput:
        """
        For each entrypoint field, specify if we should return all data, or just data that matches the coerced
        input values. Used for post-processing access request results for a given collection.

        :param pre_processed_inputs: string paths mapped to values that were used to query the current collection
        :return: FieldPaths mapped to type-coerced values that we need to match in
        access request results, or FieldPaths mapped to None if we want to return everything.

        :Example:
        owner.phone field will not be filtered but we will process the owner.identifier results to return
        values that match one of [1234, 5678, 9102]

        {FieldPath("owner", "phone"): None, FieldPath("owner", "identifier"): [1234, 5678, 9102]}
        """
        out: FieldPathNodeInput = {}
        for key, values in pre_processed_inputs.items():
            path: FieldPath = FieldPath.parse(key)
            field: Optional[Field] = self.execution_node.collection.field(path)
            if (
                field
                and path in self.execution_node.query_field_paths
                and isinstance(values, list)
            ):
                if field.return_all_elements:
                    # All data will be returned
                    out[path] = None
                else:
                    # Default behavior - we will filter values to match those in filtered
                    cast_values = [
                        field.cast(v) for v in values
                    ]  # Cast values to expected type where possible
                    filtered = list(filter(lambda x: x is not None, cast_values))
                    if filtered:
                        out[path] = filtered
        return out

    def access_results_post_processing(
        self, formatted_input_data: NodeInput, output: List[Row]
    ) -> List[Row]:
        """
        Completes post-processing filtering of access request results.

        By default, if an array field was an entry point into the node, return only array elements that *match* the
        condition.  Specifying return_all_elements = true on the field's config will instead return *all* array elements.

        Caches the data in TWO separate formats: 1) erasure format, *replaces* unmatched array elements with placeholder
        text, and 2) access request format, which *removes* unmatched array elements altogether.  If no data was filtered
        out, both cached versions will be the same.
        """
        post_processed_node_input_data: FieldPathNodeInput = (
            self.post_process_input_data(formatted_input_data)
        )

        # For erasures: cache results with non-matching array elements *replaced* with placeholder text
        placeholder_output: List[Row] = copy.deepcopy(output)
        for row in placeholder_output:
            filter_element_match(
                row, query_paths=post_processed_node_input_data, delete_elements=False
            )

        # For DSR 3.0, save data to build masking requests directly
        # on the Request Task.
        # Results saved with matching array elements preserved
        if self.request_task.id:
            self.request_task.data_for_erasures = placeholder_output

        # TODO Remove when we stop support for DSR 2.0
        # Save data to build masking requests for DSR 2.0 in Redis.
        # Results saved with matching array elements preserved
        self.resources.cache_results_with_placeholders(
            f"access_request__{self.key}", placeholder_output
        )

        # For access request results, cache results with non-matching array elements *removed*
        for row in output:
            logger.info(
                "Filtering row in {} for matching array elements.",
                self.execution_node.address,
            )
            filter_element_match(row, post_processed_node_input_data)

        if self.request_task.id:
            # Saves intermediate access results for DSR 3.0 directly on the Request Task
            self.request_task.access_data = output

        # TODO Remove when we stop support for DSR 2.0
        # Saves intermediate access results for DSR 2.0 in Redis
        self.resources.cache_object(f"access_request__{self.key}", output)

        # Return filtered rows with non-matched array data removed.
        return output

    def get_connection_config(self) -> ConnectionConfig:
        """
        Retrieves the connection configuration for the current connector.
        This method attempts to fetch the connection configuration from the database
        using the connector's configuration ID. If the configuration is found in the
        database, it is returned. Otherwise, the method returns the connector's
        current configuration.
        Returns:
            ConnectionConfig: The connection configuration object from the database
            if found, otherwise the connector's current configuration.
        """
        connection_config_id = self.connector.configuration.id

        with get_db() as db:
            connection_config = (
                db.query(ConnectionConfig)
                .filter(ConnectionConfig.id == connection_config_id)
                .first()
            )
            if connection_config:
                return connection_config

        return self.connector.configuration

    def skip_if_disabled(self) -> None:
        """Skip execution for the given collection if it is attached to a disabled ConnectionConfig."""
        connection_config: ConnectionConfig = self.get_connection_config()

        if connection_config.disabled:
            raise CollectionDisabled(
                f"Skipping collection {self.execution_node.address}. "
                f"ConnectionConfig {connection_config.key} is disabled.",
            )

    def skip_if_action_disabled(self, action_type: ActionType) -> None:
        """Skip execution for the given collection if it is attached to a ConnectionConfig that does not have the given action_type enabled."""

        # the access action is never disabled since it provides data that is needed for erasure requests
        if action_type == ActionType.access:
            return

        connection_config: ConnectionConfig = self.get_connection_config()
        if (
            connection_config.enabled_actions is not None
            and action_type not in connection_config.enabled_actions
        ):
            raise ActionDisabled(
                f"Skipping collection {self.execution_node.address}. "
                f"The {action_type} action is disabled for connection config with key '{connection_config.key}'.",
            )

    @retry(action_type=ActionType.access, default_return=[])
    def access_request(self, *inputs: List[Row]) -> List[Row]:
        """Run an access request on a single node."""
        with logger.contextualize(
            **{LoggerContextKeys.privacy_request_id.value: self.resources.request.id}
        ):
            formatted_input_data: NodeInput = self.pre_process_input_data(
                *inputs, group_dependent_fields=True
            )
            output: List[Row] = self.connector.retrieve_data(
                self.execution_node,
                self.resources.policy,
                self.resources.request,
                self.resources.privacy_request_task,
                formatted_input_data,
            )
            filtered_output: List[Row] = self.access_results_post_processing(
                self.pre_process_input_data(*inputs, group_dependent_fields=False),
                output,
            )
            self.log_end(ActionType.access)
            return filtered_output

    @retry(action_type=ActionType.erasure, default_return=0)
    def erasure_request(
        self,
        retrieved_data: List[Row],
        *erasure_prereqs: int,  # TODO Remove when we stop support for DSR 2.0. DSR 3.0 enforces with downstream_tasks.
    ) -> int:
        """Run erasure request"""

        # if there is no primary key specified in the graph node configuration
        # note this in the execution log and perform no erasures on this node
        if (
            self.connector.requires_primary_keys
            and not self.execution_node.collection.contains_field(
                lambda f: f.primary_key
            )
        ):
            logger.warning(
                'Skipping erasures on "{}" as the "{}" connector requires a primary key to be defined in one of the collection fields, but none was found.',
                self.execution_node.address,
                self.connector.configuration.connection_type,
            )
            if self.request_task.id:
                # For DSR 3.0, largely for testing. DSR 3.0 uses Request Task status
                # instead of presence of cached erasure data to know if we should rerun a node
                self.request_task.rows_masked = 0  # Saved as part of update_status
            # TODO Remove when we stop support for DSR 2.0
            self.resources.cache_erasure(self.key.value, 0)
            self.update_status(
                "No values were erased since no primary key was defined in any of the fields for this collection",
                None,
                ActionType.erasure,
                ExecutionLogStatus.complete,
            )
            return 0

        if not self.can_write_data():
            logger.warning(
                "No erasures on {} as its ConnectionConfig does not have write access.",
                self.execution_node.address,
            )
            if self.request_task.id:
                # DSR 3.0
                self.request_task.rows_masked = 0  # Saved as part of update_status
            # TODO Remove when we stop support for DSR 2.0
            self.resources.cache_erasure(self.key.value, 0)
            self.update_status(
                f"No values were erased since this connection {self.connector.configuration.key} has not been "
                f"given write access",
                None,
                ActionType.erasure,
                ExecutionLogStatus.error,
            )
            return 0

        output = self.connector.mask_data(
            self.execution_node,
            self.resources.policy,
            self.resources.request,
            self.resources.privacy_request_task,
            retrieved_data,
        )
        if self.request_task.id:
            # For DSR 3.0, largely for testing. DSR 3.0 uses Request Task status
            # instead of presence of cached erasure data to know if we should rerun a node
            self.request_task.rows_masked = (
                output  # Saved as part of update_status below
            )
        # TODO Remove when we stop support for DSR 2.0
        self.resources.cache_erasure(
            self.key.value, output
        )  # Cache that the erasure was performed in case we need to restart
        self.log_end(ActionType.erasure)
        return output

    @retry(action_type=ActionType.consent, default_return=False)
    def consent_request(self, identity: Dict[str, Any]) -> bool:
        """Run consent request request"""
        if not self.can_write_data():
            logger.warning(
                "No consent on {} as its ConnectionConfig does not have write access.",
                self.execution_node.address,
            )
            if self.request_task.id:
                # For DSR 3.0, saved as part of
                self.request_task.consent_sent = False
            self.update_status(
                f"No consent requests were sent since this connection {self.connector.configuration.key} has not been "
                f"given write access",
                None,
                ActionType.consent,
                ExecutionLogStatus.error,
            )
            return False

        output: bool = self.connector.run_consent_request(
            self.execution_node,
            self.resources.policy,
            self.resources.request,
            self.resources.privacy_request_task,
            identity,
            self.resources.session,
        )
        self.request_task.consent_sent = output
        self.log_end(ActionType.consent)
        return output


def collect_queries(
    traversal: Traversal, resources: TaskResources
) -> Dict[CollectionAddress, str]:
    """Collect all queries for dry-run"""

    def collect_queries_fn(
        tn: TraversalNode, data: Dict[CollectionAddress, str]
    ) -> None:
        if not tn.is_root_node():
            # Mock a RequestTask object in memory
            resources.privacy_request_task = tn.to_mock_request_task()
            data[tn.address] = GraphTask(resources).generate_dry_run_query()  # type: ignore

    env: Dict[CollectionAddress, str] = {}
    traversal.traverse(env, collect_queries_fn)
    return env


def start_function(seed: List[Dict[str, Any]]) -> Callable[[], List[Dict[str, Any]]]:
    """Return a function for collections with no upstream dependencies, that just start
    with seed data.

    This is used for root nodes or previously-visited nodes on restart."""

    def g() -> List[Dict[str, Any]]:
        return seed

    return g


def filter_by_enabled_actions(
    access_results: Dict[str, Any], connection_configs: List[ConnectionConfig]
) -> Dict[str, Any]:
    """Removes any access results that are associated with a connection config that doesn't have the access action enabled."""

    # create a map between the dataset and its connection config's enabled actions
    dataset_enabled_actions = {}
    for config in connection_configs:
        for dataset in config.datasets:
            dataset_enabled_actions[dataset.fides_key] = config.enabled_actions

    # use the enabled actions map to filter out the access results
    filtered_access_results = {}
    for key, value in access_results.items():
        dataset_name = key.split(":")[0]
        enabled_action = dataset_enabled_actions.get(dataset_name)
        if enabled_action is None or ActionType.access in enabled_action:
            filtered_access_results[key] = value

    return filtered_access_results


def get_cached_data_for_erasures(
    privacy_request_id: str,
) -> Dict[str, Any]:
    """
    Fetches processed access request results to be used for erasures.

    Processing may have added indicators to not mask certain elements in array data.
    """
    cache = get_cache()
    value_dict = cache.get_encoded_objects_by_prefix(
        f"PLACEHOLDER_RESULTS__{privacy_request_id}"
    )
    number_of_leading_strings_to_exclude = 3
    return {
        extract_key_for_address(k, number_of_leading_strings_to_exclude): v
        for k, v in value_dict.items()
    }


def build_affected_field_logs(
    node: ExecutionNode, policy: Policy, action_type: ActionType
) -> List[Dict[str, Any]]:
    """For a given node (collection), policy, and action_type (access or erasure) format all of the fields that
    were potentially touched to be stored in the ExecutionLogs for troubleshooting.

    :Example:
    [{
        "path": "dataset_name:collection_name:field_name",
        "field_name": "field_name",
        "data_categories": ["data_category_1", "data_category_2"]
    }]
    """

    targeted_field_paths: Dict[FieldAddress, str] = {}

    for rule in policy.rules:  # type: ignore[attr-defined]
        if rule.action_type != action_type:
            continue
        rule_categories: List[str] = rule.get_target_data_categories()
        if not rule_categories:
            continue

        collection_categories: Dict[
            str, List[FieldPath]
        ] = node.collection.field_paths_by_category  # type: ignore
        for rule_cat in rule_categories:
            for collection_cat, field_paths in collection_categories.items():
                if collection_cat.startswith(rule_cat):
                    targeted_field_paths.update(
                        {
                            node.address.field_address(field_path): collection_cat
                            for field_path in field_paths
                        }
                    )

    ret: List[Dict[str, Any]] = []
    for field_address, data_categories in targeted_field_paths.items():
        ret.append(
            {
                "path": field_address.value,
                "field_name": field_address.field_path.string_path,
                "data_categories": [data_categories],
            }
        )

    return ret


def build_consent_dataset_graph(datasets: List[DatasetConfig]) -> DatasetGraph:
    """
    Build the starting DatasetGraph for consent requests.

    Consent Graph has one node per dataset.  Nodes must be of saas type and have consent requests defined.
    """
    consent_datasets: List[GraphDataset] = []

    for dataset_config in datasets:
        connection_config: ConnectionConfig = dataset_config.connection_config

        if connection_config.connection_type != ConnectionType.saas:
            continue

        saas_config = connection_config.get_saas_config()
        if not saas_config:
            continue

        if ActionType.consent in saas_config.supported_actions:
            consent_datasets.append(
                dataset_config.get_dataset_with_stubbed_collection()
            )

    return DatasetGraph(*consent_datasets)



"""Add AWS SES email provider

Revision ID: 82883b0df5e4
Revises: c961528edfc6
Create Date: 2025-02-20 14:27:58.974466

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "82883b0df5e4"
down_revision = "c961528edfc6"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type messagingservicetype rename to messagingservicetype_old")
    op.execute(
        "create type messagingservicetype as enum('mailchimp_transactional', 'mailgun', 'twilio_text', 'twilio_email', 'aws_ses')"
    )
    op.execute(
        (
            "alter table messagingconfig alter column service_type type messagingservicetype using "
            "service_type::text::messagingservicetype"
        )
    )
    op.execute("drop type messagingservicetype_old")


def downgrade():
    op.execute("delete from messagingconfig where service_type in ('aws_ses')")
    op.execute("alter type messagingservicetype rename to messagingservicetype_old")
    op.execute(
        "create type messagingservicetype as enum('mailchimp_transactional', 'mailgun', 'twilio_text', 'twilio_email')"
    )
    op.execute(
        (
            "alter table messagingconfig alter column service_type type messagingservicetype using "
            "service_type::text::messagingservicetype"
        )
    )
    op.execute("drop type messagingservicetype_old")



"""add html format for storageconfig

Revision ID: c1885270b3cc
Revises: 5307999c0dac
Create Date: 2023-05-31 21:43:45.404454

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "c1885270b3cc"
down_revision = "5307999c0dac"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type responseformat rename to responseformat_old")
    op.execute("create type responseformat as enum('json', 'csv', 'html')")

    # drop default, change type and set default back
    op.execute("alter table storageconfig alter column format drop default")
    op.execute(
        "alter table storageconfig alter column format type responseformat using format::text::responseformat"
    )
    op.execute("alter table storageconfig alter column format set default 'json'")

    op.execute("drop type responseformat_old")


def downgrade():
    op.execute("delete from storageconfig where format = 'html'")

    op.execute("alter type responseformat rename to responseformat_old")
    op.execute("create type responseformat as enum('json', 'csv')")

    # drop default, change type and set default back
    op.execute("alter table storageconfig alter column format drop default")
    op.execute(
        "alter table storageconfig alter column format type responseformat using format::text::responseformat"
    )
    op.execute("alter table storageconfig alter column format set default 'json'")

    op.execute("drop type responseformat_old")



# pylint: disable=protected-access
from __future__ import annotations

import random
import string
from typing import TYPE_CHECKING, Any, Dict, List, Type
from uuid import uuid4

from sqlalchemy import (
    Boolean,
    Column,
    ForeignKey,
    Index,
    String,
    Text,
    UniqueConstraint,
    and_,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import RelationshipProperty, Session, relationship

from fides.api.db.base_class import Base
from fides.api.db.util import EnumColumn
from fides.api.schemas.property import PropertyType
from fides.config import get_config

# Hack to avoid circular dependency errors
if TYPE_CHECKING:
    from fides.api.models.messaging_template import MessagingTemplate
    from fides.api.models.privacy_experience import PrivacyExperienceConfig

CONFIG = get_config()


class Property(Base):
    """
    This class serves as a model for digital properties, such as websites or other online platforms.
    """

    def generate_id(self) -> str:
        """
        Generate a unique ID in the format 'FDS-XXXXXX' using uppercase alphanumeric characters.
        """
        characters = string.ascii_uppercase + string.digits
        return "FDS-" + "".join(random.choices(characters, k=6))

    @declared_attr
    def __tablename__(self) -> str:
        return "plus_property"

    id = Column(
        String,
        primary_key=True,
        default=generate_id,
    )
    # Right now, we use server default to write this val.
    # In the future we may allow ability to configure which property is the default
    is_default = Column(Boolean, server_default="f", default=False)
    name = Column(String, nullable=False, unique=True)
    type = Column(EnumColumn(PropertyType), nullable=False)
    privacy_center_config = Column(MutableDict.as_mutable(JSONB), nullable=True)
    stylesheet = Column(Text, nullable=True)

    _property_paths: RelationshipProperty[List[PropertyPath]] = relationship(
        "PropertyPath",
        cascade="all, delete-orphan",
        lazy="selectin",
    )
    paths: List[str] = association_proxy("_property_paths", "path")

    experiences: RelationshipProperty[List[PrivacyExperienceConfig]] = relationship(
        "PrivacyExperienceConfig",
        secondary="plus_privacy_experience_config_property",
        back_populates="properties",
        lazy="selectin",
    )

    messaging_templates: RelationshipProperty[List[MessagingTemplate]] = relationship(
        "MessagingTemplate",
        secondary="messaging_template_to_property",
        back_populates="properties",
        lazy="selectin",
    )

    # Only 1 property can be the default
    __table_args__ = (
        Index(
            "only_one_default",
            is_default,
            unique=True,
            postgresql_where=is_default,
        ),
    )

    @classmethod
    def create(
        cls: Type[Property],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> Property:
        experiences = data.pop("experiences", [])
        paths = data.pop("paths", None) or []

        matching_paths = PropertyPath.filter(
            db=db, conditions=(PropertyPath.path.in_(paths))
        ).all()
        if matching_paths:
            raise ValueError(
                f'The path(s) \'{", ".join([matching_path.path for matching_path in matching_paths])}\' are already associated with another property.'
            )

        # Ensure that there is at least 1 default. Relevant for new Fides users who have no properties at all.
        has_default_property = Property.get_by(db=db, field="is_default", value=True)
        if not has_default_property:
            data["is_default"] = True

        prop: Property = super().create(db=db, data=data, check_name=check_name)
        link_experience_configs_to_property(
            db, experience_configs=experiences, prop=prop
        )

        property_paths = [
            PropertyPath(property_id=prop.id, path=path) for path in set(paths)  # type: ignore
        ]
        prop._property_paths = property_paths

        return cls.persist_obj(db, prop)

    def update(self, db: Session, *, data: Dict[str, Any]) -> Property:
        experiences = data.pop("experiences", [])
        paths = data.pop("paths", None) or []

        matching_paths = (
            db.query(PropertyPath)
            .filter(
                and_(PropertyPath.path.in_(paths), PropertyPath.property_id != self.id)
            )
            .all()
        )
        if matching_paths:
            raise ValueError(
                f'The path(s) \'{", ".join([matching_path.path for matching_path in matching_paths])}\' are already associated with another property.'
            )

        # Ensure that there is at least 1 default.
        has_default_property = Property.get_by(db=db, field="is_default", value=True)
        if not has_default_property:
            data["is_default"] = True
        else:
            # Prevent the only default property from being changed to `is_default=False`
            if has_default_property.id == self.id and data.get("is_default") is False:
                raise ValueError(
                    f"Unable to set property with id {self.id} to is_default=False. You must have exactly one default property configured."
                )

        super().update(db=db, data=data)
        link_experience_configs_to_property(
            db, experience_configs=experiences, prop=self
        )

        existing_paths = {path.path: path for path in self._property_paths}
        updated_paths = set(paths)

        # delete PropertyPath instances that are not in the updated paths
        for path in set(existing_paths.keys()) - updated_paths:
            db.delete(existing_paths[path])

        # create new PropertyPath instances for paths that don't exist
        for path in updated_paths - set(existing_paths.keys()):
            self._property_paths.append(PropertyPath(property_id=self.id, path=path))

        self.save(db)
        return self


class PropertyPath(Base):
    """Table mapping url paths to properties"""

    @declared_attr
    def __tablename__(self) -> str:
        return "plus_property_path"

    property_id = Column(
        String,
        ForeignKey("plus_property.id"),
        index=True,
        nullable=False,
        primary_key=True,
        unique=False,
    )
    path = Column(String, index=True, nullable=False, unique=True)


class PrivacyExperienceConfigProperty(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "plus_privacy_experience_config_property"

    def generate_uuid(self) -> str:
        """
        Generates a uuid with a prefix based on the tablename to be used as the
        record's ID value
        """
        try:
            prefix = f"{self.current_column.table.name[:3]}_"  # type: ignore
        except AttributeError:
            prefix = ""
        uuid = str(uuid4())
        return f"{prefix}{uuid}"

    # Overrides Base.id so this is not a primary key.
    # Instead, we have a composite PK of privacy_experience_config_id and property_id
    id = Column(String(255), default=generate_uuid)

    privacy_experience_config_id = Column(
        String,
        ForeignKey("privacyexperienceconfig.id"),
        index=True,
        nullable=False,
        primary_key=True,
    )
    property_id = Column(
        String,
        ForeignKey("plus_property.id"),
        index=True,
        nullable=False,
        primary_key=True,
    )


def link_experience_configs_to_property(
    db: Session,
    *,
    experience_configs: List[Dict[str, Any]],
    prop: Property,
) -> List[PrivacyExperienceConfig]:
    """
    Link supplied experience configs to the property.
    """

    # delayed import to avoid circular declarations
    from fides.api.models.privacy_experience import PrivacyExperienceConfig

    if experience_configs:
        experience_config_ids = [
            experience_config["id"] for experience_config in experience_configs
        ]
        new_experience_configs = (
            db.query(PrivacyExperienceConfig)
            .filter(PrivacyExperienceConfig.id.in_(experience_config_ids))
            .all()
        )
        prop.experiences = new_experience_configs
    else:
        prop.experiences = []

    prop.save(db)
    return prop.experiences


class MessagingTemplateToProperty(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "messaging_template_to_property"

    messaging_template_id = Column(
        String,
        ForeignKey("messaging_template.id"),
        unique=False,
        index=True,
        nullable=False,
        primary_key=True,
    )
    property_id = Column(
        String,
        ForeignKey("plus_property.id"),
        unique=False,
        index=True,
        nullable=False,
        primary_key=True,
    )

    __table_args__ = (
        UniqueConstraint(
            "messaging_template_id",
            "property_id",
            name="messaging_template_id_property_id",
        ),
    )



"""adds webhook_id column to auditLog

Revision ID: 93af4df40cc0
Revises: 55bedede956d
Create Date: 2024-04-20 13:17:33.928813

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "93af4df40cc0"
down_revision = "55bedede956d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("auditlog", sa.Column("webhook_id", sa.String(), nullable=True))
    op.create_index(
        op.f("ix_auditlog_webhook_id"), "auditlog", ["webhook_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_auditlog_webhook_id"), table_name="auditlog")
    op.drop_column("auditlog", "webhook_id")
    # ### end Alembic commands ###



"""
This module is responsible for combining all of the different
config sections into a single config object.
"""

from functools import lru_cache
from os import getenv
from typing import Any, Dict, Optional, Tuple, Type, Union

import toml
from loguru import logger as log
from pydantic import ConfigDict, Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

from fides.common.utils import echo_red

from .admin_ui_settings import AdminUISettings
from .celery_settings import CelerySettings
from .cli_settings import CLISettings
from .consent_settings import ConsentSettings
from .credentials_settings import merge_credentials_environment
from .database_settings import DatabaseSettings
from .execution_settings import ExecutionSettings
from .fides_settings import FidesSettings
from .helpers import handle_deprecated_env_variables, handle_deprecated_fields
from .logging_settings import LoggingSettings
from .notification_settings import NotificationSettings
from .redis_settings import RedisSettings
from .security_settings import SecuritySettings
from .user_settings import UserSettings
from .utils import (
    CONFIG_KEY_ALLOWLIST,
    DEFAULT_CONFIG_PATH,
    DEFAULT_CONFIG_PATH_ENV_VAR,
    get_test_mode,
)


class FidesConfig(FidesSettings):
    """
    Composite class that encapsulates all of the config subsections
    as well as root-level values.
    """

    # Root Settings
    test_mode: bool = Field(
        default=get_test_mode(),
        description="Whether or not the application is being run in test mode.",
        exclude=True,
    )
    hot_reloading: bool = Field(
        default=getenv("FIDES__HOT_RELOAD", "").lower() == "true",
        description="Whether or not to enable hot reloading for the webserver.",
        exclude=True,
    )
    dev_mode: bool = Field(
        default=getenv("FIDES__DEV_MODE", "").lower() == "true",
        description="Similar to 'test_mode', enables certain features when true.",
        exclude=True,
    )
    oauth_instance: Optional[str] = Field(
        default=getenv("FIDES__OAUTH_INSTANCE", None),
        description="A value that is prepended to the generated 'state' param in outbound OAuth2 authorization requests. Used during OAuth2 testing to associate callback responses back to this specific Fides instance.",
        exclude=True,
    )

    # Setting Subsections
    # These should match the `settings_map` in `build_config`
    admin_ui: AdminUISettings
    consent: ConsentSettings
    cli: CLISettings
    celery: CelerySettings
    credentials: Dict = Field(
        description="This is a special section that is used to store arbitrary key/value pairs to be used as credentials."
    )
    database: DatabaseSettings
    execution: ExecutionSettings
    logging: LoggingSettings
    notifications: NotificationSettings
    redis: RedisSettings
    security: SecuritySettings
    user: UserSettings

    model_config = SettingsConfigDict(case_sensitive=True)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Set environment variables to take precedence over init values."""
        return env_settings, init_settings, file_secret_settings

    def log_all_config_values(self) -> None:
        """Output DEBUG logs of all the config values."""
        for settings in [
            self.cli,
            self.user,
            self.logging,
            self.database,
            self.notifications,
            self.redis,
            self.security,
            self.execution,
            self.admin_ui,
        ]:
            for key, value in settings.model_dump(mode="json").items():  # type: ignore
                log.debug(
                    f"Using config: {settings.Config.env_prefix}{key.upper()} = {value}",  # type: ignore
                )


def censor_config(config: Union[FidesConfig, Dict[str, Any]]) -> Dict[str, Any]:
    """
    Returns a config that is safe to expose over the API. This function will
    strip out any keys not specified in the `CONFIG_KEY_ALLOWLIST` above.
    """
    if not isinstance(config, Dict):
        as_dict = config.model_dump(mode="json")
    else:
        as_dict = config
    filtered: Dict[str, Any] = {}
    for key, value in CONFIG_KEY_ALLOWLIST.items():
        if key in as_dict:
            data = as_dict[key]
            filtered[key] = {}
            for field in value:
                if field in data:
                    filtered[key][field] = data[field]

    return filtered


def build_config(config_dict: Dict[str, Any]) -> FidesConfig:
    """
    This function builds and instantiates a FidesConfig.

    It will pull from the provided config_dict when possible for subsections,
    otherwise instantiating the settings with default values.

    The settings_map will need to be kept up-to-date manually with the
    expected subsections of the FidesConfig.
    """
    config_dict = handle_deprecated_fields(config_dict)
    config_dict = handle_deprecated_env_variables(config_dict)

    settings_map: Dict[str, Any] = {
        "admin_ui": AdminUISettings,
        "celery": CelerySettings,
        "consent": ConsentSettings,
        "cli": CLISettings,
        "database": DatabaseSettings,
        "execution": ExecutionSettings,
        "logging": LoggingSettings,
        "notifications": NotificationSettings,
        "redis": RedisSettings,
        "security": SecuritySettings,
        "user": UserSettings,
    }

    for key, value in settings_map.items():
        settings_map[key] = value.model_validate(config_dict.get(key, {}))

    # Logic for populating the user-defined credentials sub-settings.
    # this is done to allow overrides without typed pydantic models
    config_environment_dict = config_dict.get("credentials", {})
    settings_map["credentials"] = merge_credentials_environment(
        credentials_dict=config_environment_dict
    )

    fides_config = FidesConfig(**settings_map)
    return fides_config


@lru_cache(maxsize=1)
def get_config(config_path_override: str = "", verbose: bool = False) -> FidesConfig:
    """
    Attempt to load user-defined configuration file, otherwise will use defaults
    while still respecting any env vars set by the user.

    Order of file checks is:
        1. The config path provided by the user to the CLI
        2. The config path provided by the user via an env var
        3. The default config path

    This will fail if the first encountered configuration file is invalid.
    """

    env_config_path = getenv(DEFAULT_CONFIG_PATH_ENV_VAR)
    config_path = config_path_override or env_config_path or DEFAULT_CONFIG_PATH

    try:
        settings = toml.load(config_path)
        config = build_config(config_dict=settings)
        if verbose:
            print(f"> Loaded config from: {config_path}")
        return config
    except FileNotFoundError:
        pass
    except IOError:
        echo_red(f"Error reading config file: {config_path}")

    if verbose:  # pragma: no cover
        print("> Using default configuration values.")
    config = build_config(config_dict={})

    return config


def check_required_webserver_config_values(config: FidesConfig) -> None:
    """Check for required config values and print a user-friendly error message."""
    required_config_dict = {
        "security": [
            "app_encryption_key",
            "oauth_root_client_id",
            "oauth_root_client_secret",
        ]
    }

    missing_required_config_vars = []
    for subsection_key, values in required_config_dict.items():
        for key in values:
            subsection_model = dict(config).get(subsection_key, {})
            config_value = dict(subsection_model).get(key)

            if not config_value:
                missing_required_config_vars.append(".".join((subsection_key, key)))

    if missing_required_config_vars:
        echo_red(
            "\nThere are missing required configuration variables. Please add the following config variables to either the "
            "`fides.toml` file or your environment variables to start Fides: \n"
        )
        for missing_value in missing_required_config_vars:
            echo_red(f"- {missing_value}")
        echo_red(
            "\nVisit the Fides deployment documentation for more information: "
            "https://ethyca.com/docs/dev-docs/configuration/deployment"
        )

        raise SystemExit(1)


CONFIG = get_config()



from pydantic import BaseModel, field_validator

from fides.api.util.encryption.aes_gcm_encryption_scheme import verify_encryption_key
from fides.config import CONFIG


class AesEncryptionRequest(BaseModel):
    """Specifies fields provided to the AES Encryption endpoint"""

    value: str
    key: str

    @field_validator("key")
    @classmethod
    def validate_key(cls, v: str) -> bytes:
        """Convert string into bytes and verify this is the correct length"""

        key = v.encode(CONFIG.security.encoding)
        verify_encryption_key(key)
        return key


class AesEncryptionResponse(BaseModel):
    """Specifies fields returned from the AES Encryption endpoint"""

    encrypted_value: str
    nonce: str


class AesDecryptionRequest(BaseModel):
    """Specifies fields provided to the AES Decryption endpoint"""

    value: str
    key: str
    nonce: str


class AesDecryptionResponse(BaseModel):
    """Specified fields returned from the AES Decryption endpoint"""

    decrypted_value: str



from typing import Generic, Optional

from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.graph.config import DatahubClientConfig
from loguru import logger

from fides.api.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.schemas.connection_configuration.connection_secrets_datahub import (
    DatahubSchema,
)
from fides.api.service.connectors.base_connector import DB_CONNECTOR_TYPE


class DatahubConnector(Generic[DB_CONNECTOR_TYPE]):

    def __init__(self, configuration: ConnectionConfig):
        self.configuration = configuration
        self.config = DatahubSchema(**configuration.secrets or {})
        self.datahub_client = DataHubGraph(
            DatahubClientConfig(
                server=str(self.config.datahub_server_url),
                token=str(self.config.datahub_token),
            )
        )

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        logger.info(f"Testing DataHub connection for {self.configuration.key}...")
        try:
            self.datahub_client.test_connection()
            logger.info(
                f"DataHub connection test for {self.configuration.key} succeeded."
            )
            return ConnectionTestStatus.succeeded
        except Exception as e:
            logger.error(
                f"DataHub connection test for {self.configuration.key} failed: {e}"
            )
            return ConnectionTestStatus.failed



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from fides.api.schemas.saas.saas_config import Header, QueryParam, SaaSRequest
from fides.api.schemas.saas.shared_schemas import (
    ConnectorParamRef,
    DatasetRef,
    IdentityParamRef,
)


class StrategyConfiguration(BaseModel):
    """Base class for strategy configuration"""


class UnwrapPostProcessorConfiguration(StrategyConfiguration):
    """Dynamic JSON path access"""

    data_path: str


class FilterPostProcessorConfiguration(StrategyConfiguration):
    """Returns objects where a field has a given value"""

    field: str
    value: Union[str, DatasetRef, IdentityParamRef]
    exact: bool = True
    case_sensitive: bool = True


class OffsetPaginationConfiguration(StrategyConfiguration):
    """
    Increases the value of the query param `incremental_param` by the `increment_by` until the `limit` is hit
    or there is no more data available.
    """

    incremental_param: str
    increment_by: int
    limit: Optional[Union[int, ConnectorParamRef]] = None

    @field_validator("increment_by")
    @classmethod
    def check_increment_by(cls, increment_by: int) -> int:
        if increment_by == 0:
            raise ValueError("'increment_by' cannot be zero")
        if increment_by < 0:
            raise ValueError("'increment_by' cannot be negative")
        return increment_by


class LinkSource(Enum):
    """Locations where the link to the next page may be found."""

    headers = "headers"
    body = "body"


class LinkPaginationConfiguration(StrategyConfiguration):
    """Gets the URL for the next page from the headers or the body."""

    source: LinkSource
    rel: Optional[str] = None
    path: Optional[str] = None
    has_next: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        source = values.get("source")
        if source == LinkSource.headers.value and values.get("rel") is None:
            raise ValueError(
                "The 'rel' value must be specified when accessing the link from the headers"
            )
        if source == LinkSource.body.value and values.get("path") is None:
            raise ValueError(
                "The 'path' value must be specified when accessing the link from the body"
            )
        return values

    model_config = ConfigDict(use_enum_values=True)


class CursorPaginationConfiguration(StrategyConfiguration):
    """
    Extracts the cursor value from the 'field' of the last object in the array specified by 'data_path'
    """

    cursor_param: str
    field: str


class ApiKeyAuthenticationConfiguration(StrategyConfiguration):
    """
    API key parameter to be added in as a header or query param
    """

    headers: Optional[List[Header]] = None
    query_params: Optional[List[QueryParam]] = None
    body: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def validate_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        headers = values.get("headers")
        query_params = values.get("query_params")
        body = values.get("body")

        if not headers and not query_params and not body:
            raise ValueError(
                "At least one 'header', 'query_param', or 'body' object must be defined in an 'api_key' auth configuration"
            )

        return values


class BasicAuthenticationConfiguration(StrategyConfiguration):
    """
    Username and password to add basic authentication to HTTP requests
    """

    username: str
    password: Optional[str] = None


class BearerAuthenticationConfiguration(StrategyConfiguration):
    """
    Token to add as bearer authentication for HTTP requests
    """

    token: str


class QueryParamAuthenticationConfiguration(StrategyConfiguration):
    """
    Value to add as query param for HTTP requests
    """

    name: str
    value: str


class OAuth2BaseConfiguration(StrategyConfiguration):
    """
    OAuth2 endpoints for token retrieval, and token refresh.
    Includes an optional expires_in parameter (in seconds) for OAuth2 integrations that
    do not specify a TTL for the access tokens.
    """

    expires_in: Optional[int] = None
    token_request: SaaSRequest
    refresh_request: Optional[SaaSRequest] = None


class OAuth2AuthorizationCodeConfiguration(OAuth2BaseConfiguration):
    """
    Oauth Authorization that requires manual user interaction to get authorization
    The standard OAuth2 configuration but with an additional property to configure
    the authorization request for the Authorization Code flow.
    """

    authorization_request: SaaSRequest


class OAuth2ClientCredentialsConfiguration(OAuth2BaseConfiguration):
    """
    Ouath authorization that does not require manual user interation to get authorization
    The standard OAuth2 configuration, but excluding the refresh token during logging
    since the client credentials flow does not require a refresh token.
    """

    refresh_request: Optional[SaaSRequest] = Field(exclude=True)



from typing import Any, Dict, List, Optional, Union, cast

from fideslang.models import MaskingStrategies
from loguru import logger
from sqlalchemy import MetaData, Table, text
from sqlalchemy.engine import Engine
from sqlalchemy.sql import Delete, Update  # type: ignore
from sqlalchemy.sql.elements import ColumnElement, TextClause

from fides.api.graph.config import Field, FieldPath
from fides.api.graph.execution import ExecutionNode
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.namespace_meta.bigquery_namespace_meta import (
    BigQueryNamespaceMeta,
)
from fides.api.service.connectors.query_configs.query_config import (
    QueryStringWithoutTuplesOverrideQueryConfig,
)
from fides.api.util.collection_util import Row, filter_nonempty_values


class BigQueryQueryConfig(QueryStringWithoutTuplesOverrideQueryConfig):
    """
    Generates SQL valid for BigQuery
    """

    namespace_meta_schema = BigQueryNamespaceMeta

    @property
    def partitioning(self) -> Optional[Dict]:
        # Overriden from base implementation to allow for _only_ BQ partitioning, for now
        return self.node.collection.partitioning

    def get_partition_clauses(
        self,
    ) -> List[str]:
        """
        Returns the WHERE clauses specified in the partitioning spec

        Currently, only where-clause based partitioning is supported.

        TODO: derive partitions from a start/end/interval specification


        NOTE: when we deprecate `where_clause` partitioning in favor of a more proper partitioning DSL,
        we should be sure to still support the existing `where_clause` partition definition on
        any in-progress DSRs so that they can run through to completion.
        """
        partition_spec = self.partitioning
        if not partition_spec:
            logger.error(
                "Partitioning clauses cannot be retrieved, no partitioning specification found"
            )
            return []

        if where_clauses := partition_spec.get("where_clauses"):
            return where_clauses

        # TODO: implement more advanced partitioning support!

        raise ValueError(
            "`where_clauses` must be specified in partitioning specification!"
        )

    def _generate_table_name(self) -> str:
        """
        Prepends the dataset ID and project ID to the base table name
        if the BigQuery namespace meta is provided.
        """

        table_name = self.node.collection.name
        if self.namespace_meta:
            bigquery_namespace_meta = cast(BigQueryNamespaceMeta, self.namespace_meta)
            table_name = f"{bigquery_namespace_meta.dataset_id}.{table_name}"
            if project_id := bigquery_namespace_meta.project_id:
                table_name = f"{project_id}.{table_name}"
        return table_name

    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """
        Returns a query string with backtick formatting for tables that have the same names as
        BigQuery reserved words.
        """
        return f'SELECT {field_list} FROM `{self._generate_table_name()}` WHERE ({" OR ".join(clauses)})'

    def generate_masking_stmt(
        self,
        node: ExecutionNode,
        row: Row,
        policy: Policy,
        request: PrivacyRequest,
        client: Engine,
    ) -> Union[List[Update], List[Delete]]:
        """
        Generate a masking statement for BigQuery.

        If a masking override is present, it will take precedence over the policy masking strategy.
        """

        masking_override = node.collection.masking_strategy_override
        if masking_override and masking_override.strategy == MaskingStrategies.DELETE:
            logger.info(
                f"Masking override detected for collection {node.address.value}: {masking_override.strategy.value}"
            )
            return self.generate_delete(row, client)
        return self.generate_update(row, policy, request, client)

    def generate_update(
        self, row: Row, policy: Policy, request: PrivacyRequest, client: Engine
    ) -> List[Update]:
        """
        Using TextClause to insert 'None' values into BigQuery throws an exception, so we use update clause instead.
        Returns a List of SQLAlchemy Update object. Does not actually execute the update object.

        A List of multiple Update objects are returned for partitioned tables; for a non-partitioned table,
        a single Update object is returned in a List for consistent typing.

        TODO: DRY up this method and `generate_delete` a bit
        """
        update_value_map: Dict[str, Any] = self.update_value_map(row, policy, request)
        non_empty_reference_field_keys: Dict[str, Field] = filter_nonempty_values(
            {
                fpath.string_path: fld.cast(row[fpath.string_path])
                for fpath, fld in self.reference_field_paths.items()
                if fpath.string_path in row
            }
        )

        valid = len(non_empty_reference_field_keys) > 0 and update_value_map
        if not valid:
            logger.warning(
                "There is not enough data to generate a valid update statement for {}",
                self.node.address,
            )
            return []

        table = Table(self._generate_table_name(), MetaData(bind=client), autoload=True)
        where_clauses: List[ColumnElement] = [
            getattr(table.c, k) == v for k, v in non_empty_reference_field_keys.items()
        ]

        if self.partitioning:
            partition_clauses = self.get_partition_clauses()
            partitioned_queries = []
            logger.info(
                f"Generating {len(partition_clauses)} partition queries for node '{self.node.address}' in DSR execution"
            )
            for partition_clause in partition_clauses:
                partitioned_queries.append(
                    table.update()
                    .where(*(where_clauses + [text(partition_clause)]))
                    .values(**update_value_map)
                )

            return partitioned_queries

        return [table.update().where(*where_clauses).values(**update_value_map)]

    def generate_delete(self, row: Row, client: Engine) -> List[Delete]:
        """Returns a List of SQLAlchemy DELETE statements for BigQuery. Does not actually execute the delete statement.

        Used when a collection-level masking override is present and the masking strategy is DELETE.

        A List of multiple DELETE statements are returned for partitioned tables; for a non-partitioned table,
        a single DELETE statement is returned in a List for consistent typing.

        TODO: DRY up this method and `generate_update` a bit
        """

        non_empty_reference_field_keys: Dict[str, Field] = filter_nonempty_values(
            {
                fpath.string_path: fld.cast(row[fpath.string_path])
                for fpath, fld in self.reference_field_paths.items()
                if fpath.string_path in row
            }
        )

        valid = len(non_empty_reference_field_keys) > 0
        if not valid:
            logger.warning(
                "There is not enough data to generate a valid DELETE statement for {}",
                self.node.address,
            )
            return []

        table = Table(self._generate_table_name(), MetaData(bind=client), autoload=True)
        where_clauses: List[ColumnElement] = [
            getattr(table.c, k) == v for k, v in non_empty_reference_field_keys.items()
        ]

        if self.partitioning:
            partition_clauses = self.get_partition_clauses()
            partitioned_queries = []
            logger.info(
                f"Generating {len(partition_clauses)} partition queries for node '{self.node.address}' in DSR execution"
            )

            for partition_clause in partition_clauses:
                partitioned_queries.append(
                    table.delete().where(*(where_clauses + [text(partition_clause)]))
                )

            return partitioned_queries

        return [table.delete().where(*where_clauses)]

    def format_fields_for_query(
        self,
        field_paths: List[FieldPath],
    ) -> List[str]:
        """Returns field paths in a format they can be added into SQL queries.

        Only returns non-nested fields (fields with exactly one level).
        Nested fields are skipped with a warning log.
        """
        formatted_fields = []
        for field_path in field_paths:
            if len(field_path.levels) > 1:
                logger.warning(
                    f"Skipping nested field '{'.'.join(field_path.levels)}' as nested fields are not supported"
                )
            else:
                formatted_fields.append(field_path.levels[0])
        return formatted_fields

    def generate_raw_query_without_tuples(
        self, field_list: List[str], filters: Dict[str, List[Any]]
    ) -> Optional[TextClause]:
        """
        Allows executing a somewhat raw query where the field_list and filters do not depend
        on the Node or Graph structure.

        This is an override of the base class method that supports nested fields for BigQuery.

        Examples with dot-delimited field names, notice the periods are replaced with underscores in the parameter bindings:

        1. Single value filter:
           field_list = ["id", "name", "email"]
           filters = {"user.id": [123]}

           Generates: SELECT id, name, email FROM `project_id.dataset_id.table_name` WHERE (user.id = :user_id)
           With parameter binding: user_id = 123

        2. Multiple value filter:
           field_list = ["id", "name", "email"]
           filters = {"user.status": ["active", "pending"]}

           Generates: SELECT id, name, email FROM `project_id.dataset_id.table_name` WHERE (user.status IN (:user_status_in_stmt_generated_0, :user_status_in_stmt_generated_1))
           With parameter bindings: user_status_in_stmt_generated_0 = "active", user_status_in_stmt_generated_1 = "pending"
        """
        clauses = []
        query_data = {}
        for field_name, field_value in filters.items():
            # Replace dots with underscores in field names for parameter binding
            field_binding_name = field_name.replace(".", "_")
            data = set(field_value)
            if len(data) == 1:
                clauses.append(
                    self.format_clause_for_query(field_name, "=", field_binding_name)
                )
                query_data[field_binding_name] = data.pop()
            elif len(data) > 1:
                data_vals = list(data)
                query_data_keys: List[str] = []
                for val in data_vals:
                    # appending "_in_stmt_generated_" (can be any arbitrary str) so that this name has lower chance of conflicting with pre-existing column in table
                    query_data_name = (
                        field_binding_name
                        + "_in_stmt_generated_"
                        + str(data_vals.index(val))
                    )
                    query_data[query_data_name] = val
                    query_data_keys.append(self.format_query_data_name(query_data_name))
                operand = ", ".join(query_data_keys)
                clauses.append(self.format_clause_for_query(field_name, "IN", operand))

        if len(clauses) > 0:
            formatted_fields = ", ".join(field_list)
            query_str = self.get_formatted_query_string(formatted_fields, clauses)
            return text(query_str).params(query_data)

        return None



"""add fides cloud table

Revision ID: 192f23f4c968
Revises: 093bb28a8270
Create Date: 2023-09-01 17:28:22.099543

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "192f23f4c968"
down_revision = "093bb28a8270"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fidescloud",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("config", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_fidescloud_id"), "fidescloud", ["id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_fidescloud_id"), table_name="fidescloud")
    op.drop_table("fidescloud")
    # ### end Alembic commands ###



"""add s3 connection type

Revision ID: cb344673f633
Revises: 3304082a6cee
Create Date: 2024-05-31 20:46:08.829330

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cb344673f633"
down_revision = "3304082a6cee"
branch_labels = None
depends_on = None


def upgrade():
    # Add 's3' to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'manual_webhook', 'timescale', 'fides', 'sovrn', 'attentive', 'dynamodb', 'postgres', 'generic_consent_email', 'generic_erasure_email', 'scylla', 's3')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    # Remove 's3' from ConnectionType enum
    op.execute("delete from connectionconfig where connection_type in ('s3')")
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'manual_webhook', 'timescale', 'fides', 'sovrn', 'attentive', 'dynamodb', 'postgres', 'generic_consent_email', 'generic_erasure_email', 'scylla')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



from dataclasses import dataclass
from enum import Enum
from typing import Callable, Generic, TypeVar

T = TypeVar("T")


class SecretType(Enum):
    """Enum that holds all possible types of secrets across all masking strategies"""

    key = "key"
    salt = "salt"
    # the below types are used by the AES algorithm, when it calls HMAC to generate the nonce
    key_hmac = "key_hmac"
    salt_hmac = "salt_hmac"


@dataclass(unsafe_hash=True)
class MaskingSecretMeta(Generic[T]):
    """Holds metadata describing one secret"""

    masking_strategy: str
    generate_secret_func: Callable[[int], T]
    # currently length is the same for all masking secrets, but just in case we want to specify in future
    secret_length: int = 32


@dataclass
class MaskingSecretCache(Generic[T]):
    """Information required to cache a secret"""

    secret: T
    masking_strategy: str
    secret_type: SecretType



from fastapi import Depends
from sqlalchemy.orm import Session

from fides.api.api.deps import get_config, get_config_proxy, get_db
from fides.config import FidesConfig
from fides.config.config_proxy import ConfigProxy
from fides.service.dataset.dataset_config_service import DatasetConfigService
from fides.service.dataset.dataset_service import DatasetService
from fides.service.messaging.messaging_service import MessagingService
from fides.service.privacy_request.privacy_request_service import PrivacyRequestService


def get_messaging_service(
    db: Session = Depends(get_db),
    config: FidesConfig = Depends(get_config),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
) -> MessagingService:
    return MessagingService(db, config, config_proxy)


def get_privacy_request_service(
    db: Session = Depends(get_db),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
    messaging_service: MessagingService = Depends(get_messaging_service),
) -> PrivacyRequestService:
    return PrivacyRequestService(db, config_proxy, messaging_service)


def get_dataset_service(db: Session = Depends(get_db)) -> DatasetService:
    return DatasetService(db)


def get_dataset_config_service(db: Session = Depends(get_db)) -> DatasetConfigService:
    return DatasetConfigService(db)



from typing import Dict, Union
from urllib.parse import quote_plus

from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import Connection, Engine, create_engine  # type: ignore

from fides.api.graph.execution import ExecutionNode
from fides.api.schemas.connection_configuration import RedshiftSchema
from fides.api.service.connectors.query_configs.redshift_query_config import (
    RedshiftQueryConfig,
)
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.config import get_config

CONFIG = get_config()


class RedshiftConnector(SQLConnector):
    """Connector specific to Amazon Redshift"""

    secrets_schema = RedshiftSchema

    def build_ssh_uri(self, local_address: tuple) -> str:
        """Build SSH URI of format redshift+psycopg2://[user[:password]@][ssh_host][:ssh_port][/dbname]"""
        local_host, local_port = local_address

        config = self.secrets_schema(**self.configuration.secrets or {})

        port = f":{local_port}" if local_port else ""
        database = f"/{config.database}" if config.database else ""
        url = f"redshift+psycopg2://{config.user}:{config.password}@{local_host}{port}{database}"
        return url

    # Overrides BaseConnector.build_uri
    def build_uri(self) -> str:
        """Build URI of format redshift+psycopg2://user:password@[host][:port][/database]"""
        config = self.secrets_schema(**self.configuration.secrets or {})

        url_encoded_password = quote_plus(config.password)
        port = f":{config.port}" if config.port else ""
        database = f"/{config.database}" if config.database else ""
        url = f"redshift+psycopg2://{config.user}:{url_encoded_password}@{config.host}{port}{database}"
        return url

    # Overrides SQLConnector.create_client
    def create_client(self) -> Engine:
        """Returns a SQLAlchemy Engine that can be used to interact with a database"""
        connect_args: Dict[str, Union[int, str]] = {}
        connect_args["sslmode"] = "prefer"

        # keep alive settings to prevent long-running queries from causing a connection close
        connect_args["keepalives"] = 1
        connect_args["keepalives_idle"] = 30
        connect_args["keepalives_interval"] = 5
        connect_args["keepalives_count"] = 5

        if (
            self.configuration.secrets
            and self.configuration.secrets.get("ssh_required", False)
            and CONFIG.security.bastion_server_ssh_private_key
        ):
            config = self.secrets_schema(**self.configuration.secrets or {})
            self.create_ssh_tunnel(host=config.host, port=config.port)
            self.ssh_server.start()
            uri = self.build_ssh_uri(local_address=self.ssh_server.local_bind_address)
        else:
            uri = (self.configuration.secrets or {}).get("url") or self.build_uri()
        return create_engine(
            uri,
            hide_parameters=self.hide_parameters,
            echo=not self.hide_parameters,
            connect_args=connect_args,
        )

    def set_schema(self, connection: Connection) -> None:
        """Sets the search_path for the duration of the session"""
        config = self.secrets_schema(**self.configuration.secrets or {})
        if config.db_schema:
            logger.info("Setting Redshift search_path before retrieving data")
            stmt = text("SET search_path to :search_path")
            stmt = stmt.bindparams(search_path=config.db_schema)
            connection.execute(stmt)

    # Overrides SQLConnector.query_config
    def query_config(self, node: ExecutionNode) -> RedshiftQueryConfig:
        """Query wrapper corresponding to the input execution node."""
        return RedshiftQueryConfig(node)



from __future__ import annotations

from typing import Optional

from loguru import logger
from pydantic import ValidationError
from sqlalchemy import Boolean, Column, Enum, Index, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base, JSONTypeOverride
from fides.api.schemas.storage.storage import ResponseFormat, StorageType
from fides.api.schemas.storage.storage_secrets_docs_only import possible_storage_secrets
from fides.api.util.logger import Pii
from fides.api.util.storage_util import get_schema_for_secrets
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy


class StorageConfig(Base):
    """The DB ORM model for StorageConfig"""

    name = Column(String, unique=True, index=True)
    type = Column(Enum(StorageType), index=True, nullable=False)
    # allows JSON to detect in-place mutations to the structure (when used with sqlalchemy)
    details = Column(MutableDict.as_mutable(JSONB), nullable=False)
    key = Column(String, index=True, unique=True, nullable=False)
    is_default = Column(Boolean, index=True, default=False, nullable=False)
    secrets = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db

    format = Column(
        Enum(ResponseFormat), server_default="json", nullable=False, default="json"
    )

    __table_args__ = (
        Index(
            "only_one_default_per_type",
            type,
            is_default,
            unique=True,
            postgresql_where=(is_default),
        ),
    )

    def set_secrets(
        self,
        *,
        db: Session,
        storage_secrets: possible_storage_secrets,
    ) -> None:
        """Creates or updates secrets associated with a config id"""

        storage_type = self.type
        if not storage_type:
            raise ValueError("This object must have a `type` to validate secrets.")

        try:
            get_schema_for_secrets(
                storage_type=storage_type,  # type: ignore
                secrets=storage_secrets,
            )
        except (
            KeyError,
            ValidationError,
        ) as exc:
            logger.error("Error: {}", Pii(str(exc)))
            # We don't want to handle these explicitly here, only in the API view
            raise

        self.secrets = storage_secrets
        self.save(db=db)


def default_storage_config_name(storage_type: str) -> str:
    """
    Utility function for consistency in generating default storage config names.

    Returns a name to be used in a default storage config for the given type.
    """
    return f"Default Storage Config [{storage_type}]"


def get_default_storage_config_by_type(
    db: Session, storage_type: StorageType
) -> Optional[StorageConfig]:
    """
    Retrieve the default storage config for the given type by querying for the
    `StorageConfig` record with the given type and with `is_default` set to `True`.

    There should only ever be one `StorageConfig` record that
    matches this criteria at any point.

    If `local` storage type is requested and a default doesn't yet exist, we create a default.
    `local` storage configurations, at this point, don't have any configurable settings,
    so we can safely create a default here, at the time it's requested.
    """
    if not isinstance(storage_type, StorageType):
        # try to coerce into an enum
        try:
            storage_type = StorageType[storage_type]
        except KeyError:
            raise ValueError(
                "storage_type argument must be a valid StorageType enum member."
            )
    default_storage: Optional[StorageConfig] = (
        db.query(StorageConfig).filter_by(is_default=True, type=storage_type).first()
    )

    if not default_storage and storage_type == StorageType.local:
        return _create_local_default_storage(db)
    return default_storage


def _create_local_default_storage(db: Session) -> StorageConfig:
    """
    Creates a default `local` storage configuration record in the database.

    `local` storage configurations, at this point, don't have any configurable settings,
    so we can safely create a standard default here, without any user input.
    """
    return StorageConfig.create(
        db,
        data={
            "name": default_storage_config_name(StorageType.local.value),
            "type": StorageType.local,
            "is_default": True,
            "format": ResponseFormat.html,
            "details": {"naming": "request_id"},
        },
    )


def get_active_default_storage_config(db: Session) -> Optional[StorageConfig]:
    """
    Utility function to return the active default storage configuration.

    As of now, we look at the application property `storage.active_default_storage_type`
    that's been set through the application settings API to determine the active default
    storage type. We use that to then look up the default configuration for that given
    storage type.

    If no `storage.active_default_storage_type` property has been set through the settings API,
    then we fall back the default `local` storage configuration.

    TODO: Eventually, we'll want the `storage.active_default_storage_type` to be settable
    via _either_ API or env var/toml, i.e. to be properly reconciled with the global
    pydantic config module. For now, though, we just rely on this being set through API.
    """
    active_default_storage_type = (
        ConfigProxy(db).storage.active_default_storage_type or StorageType.local
    )
    return get_default_storage_config_by_type(db, active_default_storage_type)



"""add 'hidden' column to systems

Revision ID: abc53a00755b
Revises: 5a4859f74832
Create Date: 2024-11-11 20:18:27.770078

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "abc53a00755b"
down_revision = "5a4859f74832"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ctl_systems",
        sa.Column("hidden", sa.BOOLEAN(), server_default="f", nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ctl_systems", "hidden")
    # ### end Alembic commands ###



from typing import Set


class HashMigrationTracker:
    """
    Singleton to track the status of the hash migration for various models.
    """

    _migrated_models: Set[str] = set()

    @classmethod
    def is_migrated(cls, model_name: str) -> bool:
        """
        Returns True if the hash migration is complete for the given model.
        """

        return model_name in cls._migrated_models

    @classmethod
    def set_migrated(cls, model_name: str) -> None:
        """
        Sets the hash migration as complete for the given model.
        """

        cls._migrated_models.add(model_name)

    @classmethod
    def clear(cls) -> None:
        """
        Clear the tracker for testing purposes.
        """

        cls._migrated_models = set()



from typing import Any, Dict, Optional

from fastapi import Depends, Request
from fastapi.params import Security
from loguru import logger
from sqlalchemy.orm import Session
from starlette.status import HTTP_200_OK

from fides.api.api import deps
from fides.api.models.application_config import ApplicationConfig
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.application_config import (
    ApplicationConfig as ApplicationConfigSchema,
)
from fides.api.util.api_router import APIRouter
from fides.common.api import scope_registry as scopes
from fides.common.api.v1 import urn_registry as urls
from fides.config import censor_config
from fides.config import get_config as get_app_config
from fides.config.config_proxy import ConfigProxy

router = APIRouter(tags=["Config"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.CONFIG_READ])],
    response_model=Dict[str, Any],
)
def get_config(
    *, db: Session = Depends(deps.get_db), api_set: bool = False
) -> Dict[str, Any]:
    """Returns the current API exposable Fides configuration."""
    logger.info("Getting the exposable Fides configuration")
    if api_set:
        logger.info("Retrieving api-set application settings")
        return censor_config(ApplicationConfig.get_api_set(db))
    config = censor_config(get_app_config())
    return config


@router.patch(
    urls.CONFIG,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.CONFIG_UPDATE])],
    response_model=ApplicationConfigSchema,
    response_model_exclude_unset=True,
)
def patch_settings(
    *,
    db: Session = Depends(deps.get_db),
    request: Request,
    data: ApplicationConfigSchema,
) -> ApplicationConfigSchema:
    """
    Updates the global application settings record.

    Only keys provided will be updated, others will be unaffected,
    i.e. true PATCH behavior.
    """

    pruned_data = data.model_dump(exclude_none=True)
    logger.info("PATCHing application settings")
    update_config: ApplicationConfig = ApplicationConfig.update_api_set(db, pruned_data)

    ConfigProxy(db).load_current_cors_domains_into_middleware(request.app)

    return update_config.api_set


@router.put(
    urls.CONFIG,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.CONFIG_UPDATE])],
    response_model=ApplicationConfigSchema,
    response_model_exclude_unset=True,
)
def put_settings(
    *,
    db: Session = Depends(deps.get_db),
    request: Request,
    data: ApplicationConfigSchema,
) -> ApplicationConfigSchema:
    """
    Updates the global application settings record.

    The record will look exactly as it is provided, i.e. true PUT behavior.
    """
    pruned_data = data.model_dump(exclude_none=True)
    logger.info("PUTing application settings")
    update_config: ApplicationConfig = ApplicationConfig.update_api_set(
        db,
        pruned_data,
        merge_updates=False,
    )

    ConfigProxy(db).load_current_cors_domains_into_middleware(request.app)
    return update_config.api_set


@router.delete(
    urls.CONFIG,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.CONFIG_UPDATE])],
    response_model=Dict,
)
def reset_settings(
    *,
    db: Session = Depends(deps.get_db),
    request: Request,
) -> dict:
    """
    Resets the global application settings record.

    Only the "api-set" values are cleared, "config-set" values are
    not updated via any API calls
    """
    logger.info("Resetting api-set application settings")
    update_config: Optional[ApplicationConfig] = ApplicationConfig.clear_api_set(db)

    ConfigProxy(db).load_current_cors_domains_into_middleware(request.app)

    return update_config.api_set if update_config else {}



import uuid
from typing import Any, Dict, List, Optional, Union

from fideslang.validation import FidesKey
from pydantic import (
    ConfigDict,
    EmailStr,
    Field,
    StrictInt,
    StrictStr,
    ValidationError,
    field_validator,
    model_validator,
)

from fides.api.common_exceptions import BadRequest
from fides.api.custom_types import PhoneNumber
from fides.api.schemas.base_class import FidesSchema

MultiValue = Union[StrictInt, StrictStr, List[Union[StrictInt, StrictStr]]]


class IdentityBase(FidesSchema):
    """The minimum fields required to represent an identity."""

    phone_number: Optional[PhoneNumber] = None
    email: Optional[EmailStr] = None
    model_config = ConfigDict(extra="forbid")


class LabeledIdentity(FidesSchema):
    """
    An identity value with its accompanying UI label
    """

    label: str
    value: MultiValue


class Identity(IdentityBase):
    """Some PII grouping pertaining to a human"""

    # These are repeated so we can continue to forbid extra fields
    phone_number: Optional[PhoneNumber] = Field(default=None, title="Phone number")
    email: Optional[EmailStr] = Field(default=None, title="Email")
    ga_client_id: Optional[str] = Field(default=None, title="GA client ID")
    ljt_readerID: Optional[str] = Field(default=None, title="LJT reader ID")
    fides_user_device_id: Optional[str] = Field(
        default=None, title="Fides user device ID"
    )
    external_id: Optional[str] = Field(default=None, title="External ID")

    model_config = ConfigDict(extra="allow")

    def __init__(self, **data: Any):
        for field, value in data.items():
            if field not in self.model_fields:
                if isinstance(value, LabeledIdentity):
                    data[field] = value
                elif isinstance(value, dict) and "label" in value and "value" in value:
                    data[field] = LabeledIdentity(**value)
                else:
                    raise ValueError(
                        f'Custom identity "{field}" must be an instance of LabeledIdentity '
                        '(e.g. {"label": "Field label", "value": "123"})'
                    )
        super().__init__(**data)

    @field_validator("fides_user_device_id")
    @classmethod
    def validate_fides_user_device_id(cls, v: Optional[str]) -> Optional[str]:
        """Validate the uuid format of the fides user device id while still keeping the data type a string"""
        if not v:
            return v
        uuid.UUID(v, version=4)
        return v

    def dict(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """
        Returns a dictionary with LabeledIdentity values returned as simple values.
        """
        d = super().dict(*args, **kwargs)
        for key, value in d.items():
            if key in self.model_fields:
                d[key] = value
            else:
                # Turn LabeledIdentity into simple values
                # 'customer_id': {'label': 'Customer ID', 'value': '123'} -> 'customer_id': '123'
                d[key] = value.get("value")
        return d

    def model_dump(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """
        Returns a dictionary with LabeledIdentity values returned as simple values.
        In pydantic v2, model_dump is preferred over dict
        """
        d = super().model_dump(*args, **kwargs)
        for key, value in d.items():
            if key in self.model_fields:
                d[key] = value
            else:
                # Turn LabeledIdentity into simple values
                # 'customer_id': {'label': 'Customer ID', 'value': '123'} -> 'customer_id': '123'
                d[key] = value.get("value")
        return d

    def labeled_dict(
        self, include_default_labels: Optional[bool] = False
    ) -> Dict[str, Any]:
        """Returns a dictionary that preserves the labels for all custom/labeled identities."""
        d = super().model_dump()
        for field, _ in self.model_fields.items():
            value = getattr(self, field, None)
            if include_default_labels:
                d[field] = {
                    "label": self.model_fields[field].title,
                    "value": value,
                }
            else:
                d[field] = value
        for field in self.__pydantic_extra__ or {}:  # pylint:disable=not-an-iterable
            value = getattr(self, field, None)
            if isinstance(value, LabeledIdentity):
                d[field] = value.model_dump(mode="json")
            else:
                d[field] = value
        return d


class UnlabeledIdentities(FidesSchema):
    """
    A model for validating identity dictionaries where standard fields use Identity's validation
    but custom fields just need to be valued.
    """

    data: Dict[str, Any]

    @model_validator(mode="before")
    @classmethod
    def validate_identities(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if not isinstance(values, dict):
            raise BadRequest("Inputs must be JSON formatted")

        if not values:
            raise BadRequest("No inputs provided")

        standard_fields = {}
        custom_fields = {}

        # Separate standard and custom fields
        for field, value in values.items():
            if field in Identity.model_fields:
                if value is None:
                    raise BadRequest(f'Input "{field}" cannot be empty')
                standard_fields[field] = value
            else:
                if value is None:
                    raise BadRequest(f'Input "{field}" cannot be empty')
                custom_fields[field] = value

        # Validate standard fields using Identity
        try:
            Identity(**standard_fields)
        except ValidationError as e:
            error_detail = e.errors()[0]
            field = error_detail.get("loc")[0]  # type: ignore[assignment, index]
            clean_message = error_detail.get("msg")
            raise BadRequest(f'"{field}" {clean_message}')

        # Return the combined validated data
        return {"data": {**standard_fields, **custom_fields}}


class DatasetTestRequest(FidesSchema):
    """The policy key and inputs required to run a dataset test."""

    policy_key: FidesKey
    identities: UnlabeledIdentities


class CustomPrivacyRequestField(FidesSchema):
    """Schema for custom privacy request fields."""

    label: str
    # use StrictInt and StrictStr to avoid type coercion and maintain the original types
    value: MultiValue



"""connectionconfig disabled

Revision ID: c3472d75c80e
Revises: 27fe9da9d0f9
Create Date: 2022-06-08 20:15:28.924262

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c3472d75c80e"
down_revision = "27fe9da9d0f9"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "connectionconfig",
        sa.Column("disabled", sa.Boolean(), server_default="f", nullable=True),
    )
    op.add_column(
        "connectionconfig",
        sa.Column("disabled_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.execute("alter type executionlogstatus add value 'skipped'")


def downgrade():
    op.drop_column("connectionconfig", "disabled_at")
    op.drop_column("connectionconfig", "disabled")

    # Downgrade executionlogstatus
    op.execute("alter type executionlogstatus rename to executionlogstatus_old")
    op.execute(
        "create type executionlogstatus as enum('in_processing', 'pending', 'complete', 'error', 'retrying', 'paused')"
    )
    op.execute(
        (
            "alter table executionlog alter column status type executionlogstatus using "
            "status::text::executionlogstatus"
        )
    )
    op.execute("drop type executionlogstatus_old")
